import logging
import numpy as np

from numpy import tan, sin, cos
from pylops import LinearOperator

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


def zoeppritz_scattering(vp1, vs1, rho1, vp0, vs0, rho0, theta1):
    r"""Zoeppritz solution.

    Calculates the angle dependent p-wave reflectivity of an
    interface between two media for a set of incident angles.

    Parameters
    ----------
    vp1 : :obj:`float`
        P-wave velocity of the upper medium
    vs1 : :obj:`float`
        S-wave velocity of the upper medium
    rho1 : :obj:`float`
        Density of the upper medium
    vp0 : :obj:`float`
        P-wave velocity of the lower medium
    vs0 : :obj:`float`
        S-wave velocity of the lower medium
    rho0 : :obj:`float`
        Density of the lower medium
    theta1 : :obj:`np.ndarray` or :obj:`float`
        Incident angles in degrees

    Returns
    -------
    zoep : :obj:`np.ndarray`
        :math:`4 \times 4` matrix representing the scattering matrix for the
        incident angle ``theta1``

    See also
    --------
    zoeppritz_element : Single reflectivity element of Zoeppritz solution
    zoeppritz_PP : PP reflectivity element of Zoeppritz solution

    """
    # Create theta1 array of angles in radiants
    if isinstance(theta1, (int, float)):
        theta1 = np.array([float(theta1), ])
    elif isinstance(theta1, (list, tuple)):
        theta1 = np.array(theta1)
    theta1 = np.radians(theta1)

    # Set the ray parameter p
    p = sin(theta1) / vp1

    # Calculate reflection & transmission angles for Zoeppritz
    theta2 = np.arcsin(p * vp0) # Trans. angle of P-wave
    phi1 = np.arcsin(p * vs1)   # Refl. angle of converted S-wave
    phi2 = np.arcsin(p * vs0)   # Trans. angle of converted S-wave

    # Matrix form of Zoeppritz equation
    M = np.array([[-sin(theta1), -cos(phi1), sin(theta2), cos(phi2)],
                  [cos(theta1), -sin(phi1), cos(theta2), -sin(phi2)],
                  [2 * rho1 * vs1 * sin(phi1) * cos(theta1),
                   rho1 * vs1 * (1 - 2 * sin(phi1) ** 2),
                   2 * rho0 * vs0 * sin(phi2) * cos(theta2),
                   rho0 * vs0 * (1 - 2 * sin(phi2) ** 2)],
                  [-rho1 * vp1 * (1 - 2 * sin(phi1) ** 2),
                   rho1 * vs1 * sin(2 * phi1),
                   rho0 * vp0 * (1 - 2 * sin(phi2) ** 2),
                   -rho0 * vs0 * sin(2 * phi2)]], dtype='float')

    N = np.array([[sin(theta1), cos(phi1), -sin(theta2), -cos(phi2)],
                  [cos(theta1), -sin(phi1), cos(theta2), -sin(phi2)],
                  [2 * rho1 * vs1 * sin(phi1) * cos(theta1),
                   rho1 * vs1 * (1 - 2 * sin(phi1) ** 2),
                   2 * rho0 * vs0 * sin(phi2) * cos(theta2),
                   rho0 * vs0 * (1 - 2 * sin(phi2) ** 2)],
                  [rho1 * vp1 * (1 - 2 * sin(phi1) ** 2),
                   -rho1 * vs1 * sin(2 * phi1),
                   - rho0 * vp0 * (1 - 2 * sin(phi2) ** 2),
                   rho0 * vs0 * sin(2 * phi2)]], dtype='float')

    # Create Zoeppritz coefficient for all angles
    zoep = np.zeros((4, 4, M.shape[-1]))
    for i in range(M.shape[-1]):
        Mi = M[..., i]
        Ni = N[..., i]
        dt = np.dot(np.linalg.inv(Mi), Ni)
        zoep[..., i] = dt

    return zoep


def zoeppritz_element(vp1, vs1, rho1, vp0, vs0, rho0, theta1, element='PdPu'):
    """Single element of Zoeppritz solution.

    Simple wrapper to :py:class:`pylops.avo.avo.scattering_matrix`,
    returning any mode reflection coefficient from the Zoeppritz
    scattering matrix for specific combination of incident
    and reflected wave and a set of incident angles

    Parameters
    ----------
    vp1 : :obj:`float`
        P-wave velocity of the upper medium
    vs1 : :obj:`float`
        S-wave velocity of the upper medium
    rho1 : :obj:`float`
        Density of the upper medium
    vp0 : :obj:`float`
        P-wave velocity of the lower medium
    vs0 : :obj:`float`
        S-wave velocity of the lower medium
    rho0 : :obj:`float`
        Density of the lower medium
    theta1 : :obj:`np.ndarray` or :obj:`float`
        Incident angles in degrees
    element : :obj:`str`, optional
        specific choice of incident and reflected wave combining
        any two of the following strings: ``Pd`` P-wave downgoing,
        ``Sd`` S-wave downgoing, ``Pu`` P-wave upgoing,
        ``Su`` S-wave upgoing (e.g., ``PdPu``)

    Returns
    -------
    refl : :obj:`np.ndarray`
        reflectivity values for all input angles for specific combination
        of incident and reflected wave.

    See also
    --------
    zoeppritz_scattering : Zoeppritz solution
    zoeppritz_PP : PP reflectivity element of Zoeppritz solution

    """
    elements = np.array([['PdPu', 'SdPu', 'PuPu', 'SuPu'],
                         ['PdSu', 'SdSu', 'PuSu', 'SuSu'],
                         ['PdPd', 'SdPd', 'PuPd', 'SuPd'],
                         ['PdSd', 'SdSd', 'PuSd', 'SuSd']])
    refl = zoeppritz_scattering(vp1, vs1, rho1, vp0, vs0, rho0, theta1)
    element = np.where(elements == element)
    return np.squeeze(refl[element])


def zoeppritz_pp(vp1, vs1, rho1, vp0, vs0, rho0, theta1):
    """PP reflection coefficient from the Zoeppritz scattering matrix.

    Simple wrapper to :py:class:`pylops.avo.avo.scattering_matrix`,
    returning the PP reflection coefficient from the Zoeppritz
    scattering matrix for a set of incident angles

    Parameters
    ----------
    vp1 : :obj:`float`
        P-wave velocity of the upper medium
    vs1 : :obj:`float`
        S-wave velocity of the upper medium
    rho1 : :obj:`float`
        Density of the upper medium
    vp0 : :obj:`float`
        P-wave velocity of the lower medium
    vs0 : :obj:`float`
        S-wave velocity of the lower medium
    rho0 : :obj:`float`
        Density of the lower medium
    theta1 : :obj:`np.ndarray` or :obj:`float`
        Incident angles in degrees

    Returns
    -------
    PPrefl : :obj:`np.ndarray`
        PP reflectivity values for all input angles.

    See also
    --------
    zoeppritz_scattering : Zoeppritz solution
    zoeppritz_element : Single reflectivity element of Zoeppritz solution

    """
    PPrefl = zoeppritz_element(vp1, vs1, rho1, vp0, vs0, rho0, theta1, 'PdPu')
    return PPrefl


def approx_zoeppritz_pp(vp1, vs1, rho1, vp0, vs0, rho0, theta1):
    """PP reflection coefficient from the approximate Zoeppritz equation.

    Approximate calculation of PP reflection from the Zoeppritz
    scattering matrix for a set of incident angles [1]_.

    .. [1] Dvorkin et al. Seismic Reflections of Rock Properties.
       Cambridge. 2014.

    Parameters
    ----------
    vp1 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        P-wave velocity of the upper medium
    vs1 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        S-wave velocity of the upper medium
    rho1 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        Density of the upper medium
    vp0 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        P-wave velocity of the lower medium
    vs0 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        S-wave velocity of the lower medium
    rho0 : :obj:`np.ndarray` or :obj:`list` or :obj:`tuple`
        Density of the lower medium
    theta1 : :obj:`np.ndarray` or :obj:`float`
        Incident angles in degrees

    Returns
    -------
    PPrefl : :obj:`np.ndarray`
        PP reflectivity values for all input angles.

    See also
    --------
    zoeppritz_scattering : Zoeppritz solution
    zoeppritz_element : Single reflectivity element of Zoeppritz solution
    zoeppritz_pp : PP reflectivity element of Zoeppritz solution

    """
    vp1, vs1, rho1 = np.array(vp1), np.array(vs1), np.array(rho1)
    vp0, vs0, rho0 = np.array(vp0), np.array(vs0), np.array(rho0)

    # Incident P
    theta1 = theta1[:, np.newaxis] if vp1.size > 1 else theta1
    theta1 = np.deg2rad(theta1)

    # Ray parameter and reflected P
    p = np.sin(theta1) / vp1
    theta0 = np.arcsin(p * vp0)

    # Reflected S
    phi1 = np.arcsin(p * vs1)
    # Transmitted S
    phi0 = np.arcsin(p * vs0)

    # Coefficients
    a = rho0 * (1 - 2 * np.sin(phi0)**2.) - rho1 * (1 - 2 * np.sin(phi1)**2.)
    b = rho0 * (1 - 2 * np.sin(phi0)**2.) + 2 * rho1 * np.sin(phi1)**2.
    c = rho1 * (1 - 2 * np.sin(phi1)**2.) + 2 * rho0 * np.sin(phi0)**2.
    d = 2 * (rho0 * vs0**2 - rho1 * vs1**2)

    E = (b * np.cos(theta1) / vp1) + (c * np.cos(theta0) / vp0)
    F = (b * np.cos(phi1) / vs1) + (c * np.cos(phi0) / vs0)
    G = a - d * np.cos(theta1)/vp1 * np.cos(phi0)/vs0
    H = a - d * np.cos(theta0)/vp0 * np.cos(phi1)/vs1

    D = E*F + G*H*p**2

    rpp = (1 / D) * (F * (b * (np.cos(theta1) / vp1) - c *
                          (np.cos(theta0) / vp0)) -
                     H * p ** 2 * (a + d * (np.cos(theta1) / vp1) *
                                   (np.cos(phi0) / vs0)))

    return rpp


def akirichards(theta, vsvp, n=1):
    r"""Three terms Aki-Richards approximation.

    Computes the coefficients of the of three terms Aki-Richards approximation
    for a set of angles and a constant or variable VS/VP ratio.

    Parameters
    ----------
    theta : :obj:`np.ndarray`
        Incident angles in degrees
    vsvp : :obj:`np.ndarray` or :obj:`float`
        VS/VP ratio
    n : :obj:`int`, optional
        number of samples (if ``vsvp`` is a scalare)

    Returns
    -------
    G1 : :obj:`np.ndarray`
        first coefficient of three terms Aki-Richards approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`
    G2 : :obj:`np.ndarray`
        second coefficient of three terms Aki-Richards approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`
    G3 : :obj:`np.ndarray`
        third coefficient of three terms Aki-Richards approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`

    Notes
    -----
    The three terms Aki-Richards approximation is used to compute the
    reflection coefficient as linear combination of contrasts in
    :math:`V_P`, :math:`V_S`, and :math:`\rho`. More specifically:

    .. math::
        R(\theta) = G_1(\theta) \frac{\Delta V_P}{\bar{V_P}} + G_2(\theta)
        \frac{\Delta V_S}{\bar{V_S}} + G_3(\theta)
        \frac{\Delta \rho}{\bar{\rho}}

    where :math:`G_1(\theta) = \frac{1}{2 cos^2 \theta}`,
    :math:`G_2(\theta) = -4 (V_S/V_P)^2 sin^2 \theta`,
    :math:`G_3(\theta) = 0.5 - 2 (V_S/V_P)^2 sin^2 \theta`,
    :math:`\frac{\Delta V_P}{\bar{V_P}} = 2 \frac{V_{P,2}-V_{P,1}}{V_{P,2}+V_{P,1}}`,
    :math:`\frac{\Delta V_S}{\bar{V_S}} = 2 \frac{V_{S,2}-V_{S,1}}{V_{S,2}+V_{S,1}}`, and
    :math:`\frac{\Delta \rho}{\bar{\rho}} = 2 \frac{\rho_2-\rho_1}{\rho_2+\rho_1}`.

    """
    theta = np.deg2rad(theta)
    vsvp = vsvp*np.ones(n) if not isinstance(vsvp, np.ndarray) else vsvp

    theta = theta[:, np.newaxis] if vsvp.size > 1 else theta
    vsvp = vsvp[:, np.newaxis].T if vsvp.size > 1 else vsvp

    G1 = 1. / (2. * cos(theta) ** 2) + 0 * vsvp
    G2 = -4. * vsvp ** 2 * np.sin(theta) ** 2
    G3 = 0.5 - 2. * vsvp ** 2 * sin(theta) ** 2

    return G1, G2, G3


def fatti(theta, vsvp, n=1):
    r"""Three terms Fatti approximation.

    Computes the coefficients of the three terms Fatti approximation
    for a set of angles and a constant or variable VS/VP ratio.

    Parameters
    ----------
    theta : :obj:`np.ndarray`
        Incident angles in degrees
    vsvp : :obj:`np.ndarray` or :obj:`float`
        VS/VP ratio
    n : :obj:`int`, optional
        number of samples (if ``vsvp`` is a scalare)

    Returns
    -------
    G1 : :obj:`np.ndarray`
        first coefficient of three terms Smith-Gidlow approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`
    G2 : :obj:`np.ndarray`
        second coefficient of three terms Smith-Gidlow approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`
    G3 : :obj:`np.ndarray`
        third coefficient of three terms Smith-Gidlow approximation
        :math:`[n_{theta}  \times  n_{vsvp}]`

    Notes
    -----
    The three terms Fatti approximation is used to compute the reflection
    coefficient as linear combination of contrasts in :math:`AI`,
    :math:`SI`, and :math:`\rho`. More specifically:

    .. math::
        R(\theta) = G_1(\theta) \frac{\Delta AI}{\bar{AI}} + G_2(\theta)
        \frac{\Delta SI}{\bar{SI}} +
        G_3(\theta) \frac{\Delta \rho}{\bar{\rho}}

    where :math:`G_1(\theta) = 0.5 (1 + tan^2 \theta)`,
    :math:`G_2(\theta) = -4 (V_S/V_P)^2 sin^2 \theta`,
    :math:`G_3(\theta) = 0.5 (4 (V_S/V_P)^2 sin^2 \theta - tan^2 \theta)`,
    :math:`\frac{\Delta AI}{\bar{AI}} = 2 \frac{AI_2-AI_1}{AI_2+AI_1}`.
    :math:`\frac{\Delta SI}{\bar{SI}} = 2 \frac{SI_2-SI_1}{SI_2+SI_1}`.
    :math:`\frac{\Delta \rho}{\bar{\rho}} = 2 \frac{\rho_2-\rho_1}{\rho_2+\rho_1}`.

    """
    theta = np.deg2rad(theta)
    vsvp = vsvp*np.ones(n) if not isinstance(vsvp, np.ndarray) else vsvp

    theta = theta[:, np.newaxis] if vsvp.size > 1 else theta
    vsvp = vsvp[:, np.newaxis].T if vsvp.size > 1 else vsvp

    G1 = 0.5 * (1 + np.tan(theta) ** 2) + 0 * vsvp
    G2 = -4 * vsvp ** 2 * np.sin(theta)** 2
    G3 = 0.5 * (4 * vsvp ** 2 * np.sin(theta) ** 2 - tan(theta) ** 2)

    return G1, G2, G3


class AVOLinearModelling(LinearOperator):
    r"""AVO Linearized modelling.

    Create operator to be applied to a combination of elastic parameters
    for generation of seismic pre-stack reflectivity.

    Parameters
    ----------
    theta : :obj:`np.ndarray`
        Incident angles in degrees
    vsvp : :obj:`np.ndarray` or :obj:`float`
        VS/VP ratio
    nt0 : :obj:`int`, optional
        number of samples (if ``vsvp`` is a scalar)
    spatdims : :obj:`int` or :obj:`tuple`, optional
        Number of samples along spatial axis (or axes)
        (``None`` if only one dimension is available)
    linearization : :obj:`str`, optional
        choice of linearization, ``akirich``: Aki-Richards,
        ``fatti``: Fatti
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)

    Raises
    ------
    NotImplementedError
        If ``linearization`` is not an implemented linearization

    Notes
    -----
    The AVO linearized operator performs a linear combination of three
    (or two) elastic parameters arranged in input vector :math:`\mathbf{m}`
    of size :math:`n_{t0} \times N` to create the so-called seismic
    reflectivity:

    .. math::
        r(t, \theta, x, y) = \sum_{i=1}^N G_i(t, \theta) m_i(t, x, y) \qquad
        \forall \quad t, \theta

    where :math:`N=2/3`. Note that the reflectivity can be in 1d, 2d or 3d
    and ``spatdims`` contains the dimensions of the spatial axis (or axes)
    :math:`x` and :math:`y`.

    """
    def __init__(self, theta, vsvp=0.5, nt0=1, spatdims=None,
                 linearization='akirich', dtype='float64'):
        self.nt0 = nt0 if not isinstance(vsvp, np.ndarray) else len(vsvp)
        self.ntheta = len(theta)
        if spatdims is None:
            self.spatdims = ()
            nspatdims = 1
        else:
            self.spatdims = spatdims if isinstance(spatdims, tuple) \
                else (spatdims,)
            nspatdims = np.prod(spatdims)

        # Compute AVO coefficients
        if linearization == 'akirich':
            Gs = akirichards(theta, vsvp, n=self.nt0)
        elif linearization == 'fatti':
            Gs = fatti(theta, vsvp, n=self.nt0)
        else:
            logging.error('%s not an available '
                          'linearization...', linearization)
            raise NotImplementedError('%s not an available linearization...'
                                      % linearization)

        self.G = np.concatenate([gs.T[:, np.newaxis] for gs in Gs], axis=1)
        # add dimensions to G to account for horizonal axes
        for _ in range(len(self.spatdims)):
            self.G = self.G[...,np.newaxis]
        self.npars = len(Gs)
        self.shape = (self.nt0*self.ntheta*nspatdims,
                      self.nt0*self.npars*nspatdims)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if self.spatdims is None:
            x = x.reshape(self.nt0, self.npars)
        else:
            x = x.reshape((self.nt0, self.npars,) + self.spatdims)
        y = np.sum(self.G * x[:, :, np.newaxis], axis=1)
        return y

    def _rmatvec(self, x):
        if self.spatdims is None:
            x = x.reshape(self.nt0, self.ntheta)
        else:
            x = x.reshape((self.nt0, self.ntheta,) + self.spatdims)
        y = np.sum(self.G * x[:, np.newaxis], axis=2)
        return y



import numpy as np
from pylops import LinearOperator


class SecondDerivative(LinearOperator):
    r"""Second derivative.

    Apply second-order second derivative.

    Parameters
    ----------
    N : :obj:`int`
        Number of samples in model.
    dims : :obj:`tuple`, optional
        Number of samples for each dimension
        (``None`` if only one dimension is available)
    dir : :obj:`int`, optional
        Direction along which smoothing is applied.
    sampling : :obj:`float`, optional
        Sampling step ``dx``.
    edge : :obj:`bool`, optional
        Use reduced order derivative at edges (``True``) or
        ignore them (``False``)
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    The SecondDerivative operator applies a second derivative to any chosen
    direction of a multi-dimensional array.

    For simplicity, given a one dimensional array, the second-order centered
    first derivative is:

    .. math::
        y[i] = (x[i+1] - 2x[i] + x[i-1]) / dx^2

    """
    def __init__(self, N, dims=None, dir=0, sampling=1,
                 edge=False, dtype='float64'):
        self.N = N
        self.sampling = sampling
        self.edge = edge
        if dims is None:
            self.dims = (self.N, )
            self.reshape = False
        else:
            if np.prod(dims) != self.N:
                raise ValueError('product of dims must equal N!')
            else:
                self.dims = dims
                self.reshape = True
        self.dir = dir if dir >= 0 else len(self.dims) + dir
        self.shape = (self.N, self.N)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if not self.reshape:
            x = x.squeeze()
            y = np.zeros(self.N, self.dtype)
            y[1:-1] = (x[2:] - 2*x[1:-1] + x[0:-2]) / self.sampling**2
            if self.edge:
                y[0] = (x[0] - 2*x[1] + x[2]) / self.sampling**2
                y[-1] = (x[-3] - 2*x[-2] + x[-1]) / self.sampling**2
        else:
            x = np.reshape(x, self.dims)
            if self.dir > 0:  # need to bring the dim. to derive to first dim.
                x = np.swapaxes(x, self.dir, 0)
            y = np.zeros(x.shape, self.dtype)
            y[1:-1] = (x[2:] - 2*x[1:-1] + x[0:-2])/self.sampling**2
            if self.edge:
                y[0] = (x[0] - 2*x[1] + x[2]) / self.sampling ** 2
                y[-1] = (x[-3] - 2*x[-2] + x[-1]) / self.sampling ** 2
            if self.dir > 0:
                y = np.swapaxes(y, 0, self.dir)
            y = y.ravel()
        return y

    def _rmatvec(self, x):
        if not self.reshape:
            x = x.squeeze()
            y = np.zeros(self.N, self.dtype)
            y[0:-2] += (x[1:-1]) / self.sampling**2
            y[1:-1] -= (2*x[1:-1]) / self.sampling**2
            y[2:] += (x[1:-1]) / self.sampling**2
            if self.edge:
                y[0] += x[0] / self.sampling**2
                y[1] -= 2 * x[0] / self.sampling**2
                y[2] += x[0] / self.sampling ** 2
                y[-3] += x[-1] / self.sampling**2
                y[-2] -= 2 * x[-1] / self.sampling**2
                y[-1] += x[-1] / self.sampling**2
        else:
            x = np.reshape(x, self.dims)
            if self.dir > 0:  # need to bring the dim. to derive to first dim.
                x = np.swapaxes(x, self.dir, 0)
            y = np.zeros(x.shape, self.dtype)
            y[0:-2] += (x[1:-1]) / self.sampling**2
            y[1:-1] -= (2*x[1:-1]) / self.sampling**2
            y[2:] += (x[1:-1]) / self.sampling**2
            if self.edge:
                y[0] += x[0] / self.sampling ** 2
                y[1] -= 2 * x[0] / self.sampling ** 2
                y[2] += x[0] / self.sampling ** 2
                y[-3] += x[-1] / self.sampling ** 2
                y[-2] -= 2 * x[-1] / self.sampling ** 2
                y[-1] += x[-1] / self.sampling ** 2
            if self.dir > 0:
                y = np.swapaxes(y, 0, self.dir)
            y = y.ravel()
        return y



import numpy as np
import numpy.ma as np_ma

from pylops import LinearOperator


class Restriction(LinearOperator):
    r"""Restriction (or sampling) operator.

    Extract subset of values from input vector at locations ``iava``
    in forward mode and place those values at locations ``iava``
    in an otherwise zero vector in adjoint mode.

    Parameters
    ----------
    M : :obj:`int`
        Number of samples in model.
    iava : :obj:`list` or :obj:`numpy.ndarray`
        Integer indices of available samples for data selection.
    dims : :obj:`list`
        Number of samples for each dimension
        (``None`` if only one dimension is available)
    dir : :obj:`int`, optional
        Direction along which restriction is applied.
    dtype : :obj:`str`, optional
        Type of elements in input array.
    inplace : :obj:`bool`, optional
        Work inplace (``True``) or make a new copy (``False``). By default,
        data is a reference to the model (in forward) and model is a reference
        to the data (in adjoint).

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved
        explicitly (``True``) or not (``False``)

    See Also
    --------
    pylops.signalprocessing.Interp : Interpolation operator

    Notes
    -----
    Extraction (or *sampling*) of a subset of :math:`N` values at locations
    ``iava`` from an input (or model) vector :math:`\mathbf{x}` of size
    :math:`M` can be expressed as:

    .. math::

        y_i = x_{l_i}  \quad \forall i=1,2,...,N

    where :math:`\mathbf{l}=[l_1, l_2,..., l_N]` is a vector containing the indeces
    of the original array at which samples are taken.

    Conversely, in adjoint mode the available values in the data vector
    :math:`\mathbf{y}` are placed at locations
    :math:`\mathbf{l}=[l_1, l_2,..., l_M]` in the model vector:

    .. math::

        x_{l_i} = y_i  \quad \forall i=1,2,...,N

    and :math:`x_{j}=0 j \neq l_i` (i.e., at all other locations in input
    vector).

    """
    def __init__(self, M, iava, dims=None, dir=0,
                 dtype='float64', inplace=True):
        self.M = M
        self.dir = dir
        self.iava = iava
        if dims is None:
            self.N = len(iava)
            self.dims = (self.M, )
            self.reshape = False
        else:
            if np.prod(dims) != self.M:
                raise ValueError('product of dims must equal M!')
            else:
                self.dims = dims # model dimensions
                self.dimsd = list(dims) # data dimensions
                self.dimsd[self.dir] = len(iava)
                self.iavareshape = [1] * self.dir + [len(self.iava)] + \
                                   [1] * (len(self.dims) - self.dir - 1)
                self.N = np.prod(self.dimsd)
                self.reshape = True
        self.inplace = inplace
        self.shape = (self.N, self.M)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if not self.inplace: x = x.copy()
        if not self.reshape:
            y = x[self.iava]
        else:
            x = np.reshape(x, self.dims)
            y = np.take(x, self.iava, axis=self.dir)
            y = y.ravel()
        return y

    def _rmatvec(self, x):
        if not self.inplace: x = x.copy()
        if not self.reshape:
            y = np.zeros(self.dims, dtype=self.dtype)
            y[self.iava] = x
        else:
            x = np.reshape(x, self.dimsd)
            y = np.zeros(self.dims, dtype=self.dtype)
            np.put_along_axis(y, np.reshape(self.iava, self.iavareshape),
                              x, axis=self.dir)
            y = y.ravel()
        return y

    def mask(self, x):
        """Apply mask to input signal returning a signal of same size with
        values at ``iava`` locations and ``0`` at other locations

        Parameters
        ----------
        x : :obj:`numpy.ndarray`
            Input array (can be either flattened or not)

        Returns
        ----------
        y : :obj:`numpy.ma.core.MaskedArray`
            Masked array.

        """
        y = np_ma.array(np.zeros(self.dims), mask=np.ones(self.dims),
                        dtype=self.dtype)
        if self.reshape:
            x = np.reshape(x, self.dims)
            x = np.swapaxes(x, self.dir, 0)
            y = np.swapaxes(y, self.dir, 0)
        y.mask[self.iava] = False
        y[self.iava] = x[self.iava]
        if self.reshape:
            y = np.swapaxes(y, 0, self.dir)
        return y



from pylops import LinearOperator
from pylops.basicoperators import Gradient, Diagonal, Sum


def FirstDirectionalDerivative(dims, v, sampling=1, edge=False,
                               dtype='float64', kind='centered'):
    r"""First Directional derivative.

    Apply directional derivative operator to a multi-dimensional
    array (at least 2 dimensions are required) along either a single common
    direction or different directions for each point of the array.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension.
    v : :obj:`np.ndarray`, optional
        Single direction (array of size :math:`n_{dims}`) or group of directions
        (array of size :math:`[n_{dims} \times n_{d0} \times ... \times n_{d_{n_{dims}}}]`)
    sampling : :obj:`tuple`, optional
        Sampling steps for each direction.
    edge : :obj:`bool`, optional
        Use reduced order derivative at edges (``True``) or
        ignore them (``False``).
    dtype : :obj:`str`, optional
        Type of elements in input array.
    kind : :obj:`str`, optional
        Derivative kind (``forward``, ``centered``, or ``backward``).

    Returns
    -------
    ddop : :obj:`pylops.LinearOperator`
        First directional derivative linear operator

    Notes
    -----
    The FirstDirectionalDerivative applies a first-order derivative
    to a multi-dimensional array along the direction defined by the unitary
    vector :math:`\mathbf{v}`:

    .. math::
        df_\mathbf{v} =
            \nabla f \mathbf{v}

    or along the directions defined by the unitary vectors
    :math:`\mathbf{v}(x, y)`:

    .. math::
        df_\mathbf{v}(x,y) =
            \nabla f(x,y) \mathbf{v}(x,y)

    where we have here considered the 2-dimensional case.

    This operator can be easily implemented as the concatenation of the
    :py:class:`pylops.Gradient` operator and the :py:class:`pylops.Diagonal`
    operator with :math:`\mathbf{v}` along the main diagonal.

    """
    Gop = Gradient(dims, sampling=sampling, edge=edge, kind=kind, dtype=dtype)
    if v.ndim == 1:
        Dop = Diagonal(v, dims=[len(dims)]+list(dims), dir=0, dtype=dtype)
    else:
        Dop = Diagonal(v.ravel(), dtype=dtype)
    Sop = Sum(dims=[len(dims)]+list(dims), dir=0, dtype=dtype)
    ddop = Sop * Dop * Gop
    return LinearOperator(ddop)


def SecondDirectionalDerivative(dims, v, sampling=1, edge=False,
                                dtype='float64'):
    r"""Second Directional derivative.

    Apply second directional derivative operator to a multi-dimensional
    array (at least 2 dimensions are required) along either a single common
    direction or different directions for each point of the array.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension.
    v : :obj:`np.ndarray`, optional
        Single direction (array of size :math:`n_{dims}`) or group of directions
        (array of size :math:`[n_{dims} \times n_{d0} \times ... \times n_{d_{n_{dims}}}]`)
    sampling : :obj:`tuple`, optional
        Sampling steps for each direction.
    edge : :obj:`bool`, optional
        Use reduced order derivative at edges (``True``) or
        ignore them (``False``).
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Returns
    -------
    ddop : :obj:`pylops.LinearOperator`
        Second directional derivative linear operator

    Notes
    -----
    The SecondDirectionalDerivative applies a second-order derivative
    to a multi-dimensional array along the direction defined by the unitary
    vector :math:`\mathbf{v}`:

    .. math::
        d^2f_\mathbf{v} =
            - D_\mathbf{v}^T [D_\mathbf{v} f]

    where :math:`D_\mathbf{v}` is the first-order directional derivative
    implemented by :func:`pylops.SecondDirectionalDerivative`.

    This operator is sometimes also referred to as directional Laplacian
    in the literature.
    """
    Dop = FirstDirectionalDerivative(dims, v, sampling=sampling,
                                     edge=edge, dtype=dtype)
    ddop = - Dop.H * Dop
    return LinearOperator(ddop)



import logging
import warnings
import numpy as np

from scipy.sparse.linalg import lsqr
from scipy.signal import filtfilt
from scipy.ndimage.filters import convolve1d as sp_convolve1d

from pylops import Diagonal, Identity, Transpose
from pylops.signalprocessing import FFT, Fredholm1
from pylops.utils import dottest as Dottest
from pylops.optimization.leastsquares import PreconditionedInversion

#logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


def _MDC(G, nt, nv, dt=1., dr=1., twosided=True, fast=None, dtype=None,
         transpose=True, saveGt=True, conj=False, prescaled=False,
         _Identity=Identity, _Transpose=Transpose, _FFT=FFT,
         _Fredholm1=Fredholm1, args_Identity={}, args_Transpose={},
         args_FFT={}, args_Identity1={}, args_Transpose1={},
         args_FFT1={}, args_Fredholm1={}):
    r"""Multi-dimensional convolution.

    Used to be able to provide operators from different libraries to
    MDC. It operates in the same way as public method
    (PoststackLinearModelling) but has additional input parameters allowing
    passing a different operator and additional arguments to be passed to such
    operator.

    """
    warnings.warn('A new implementation of MDC is provided in v1.5.0. This '
                  'currently affects only the inner working of the operator, '
                  'end-users can continue using the operator in the same way. '
                  'Nevertheless, it is now recommended to start using the '
                  'operator with transpose=True, as this behaviour will '
                  'become default in version v2.0.0 and the behaviour with '
                  'transpose=False will be deprecated.', FutureWarning)

    if twosided and nt % 2 == 0:
        raise ValueError('nt must be odd number')

    # transpose G
    if transpose:
        G = np.transpose(G, axes=(2, 0, 1))

    # find out dtype of G
    dtype = G[0, 0, 0].dtype
    rdtype = np.real(np.ones(1, dtype=dtype)).dtype

    # create Fredholm operator
    if prescaled:
        Frop = _Fredholm1(G, nv, saveGt=saveGt,
                          dtype=dtype, **args_Fredholm1)
    else:
        Frop = _Fredholm1(dr * dt * np.sqrt(nt) * G, nv, saveGt=saveGt,
                          dtype=dtype, **args_Fredholm1)
    if conj:
        Frop = Frop.conj()

    # create FFT operators
    nfmax, ns, nr = G.shape
    # ensure that nfmax is not bigger than allowed
    nfft = int(np.ceil((nt + 1) / 2))
    if nfmax > nfft:
        nfmax = nfft
        logging.warning('nfmax set equal to ceil[(nt+1)/2=%d]' % nfmax)

    Fop = _FFT(dims=(nt, nr, nv), dir=0, real=True,
               fftshift=twosided, dtype=rdtype, **args_FFT)
    F1op = _FFT(dims=(nt, ns, nv), dir=0, real=True,
                fftshift=False, dtype=rdtype, **args_FFT1)

    # create Identity operator to extract only relevant frequencies
    Iop = _Identity(N=nfmax * nr * nv, M=nfft * nr * nv,
                    inplace=True, dtype=dtype, **args_Identity)
    I1op = _Identity(N=nfmax * ns * nv, M=nfft * ns * nv,
                     inplace=True, dtype=dtype, **args_Identity1)
    F1opH = F1op.H
    I1opH = I1op.H

    # create transpose operator
    if transpose:
        dims = [nr, nt] if nv == 1 else [nr, nv, nt]
        axes = (1, 0) if nv == 1 else (2, 0, 1)
        Top = _Transpose(dims, axes, dtype=dtype, **args_Transpose)

        dims = [nt, ns] if nv == 1 else [nt, ns, nv]
        axes = (1, 0) if nv == 1 else (1, 2, 0)
        TopH = _Transpose(dims, axes, dtype=dtype, **args_Transpose1)

    # create MDC operator
    MDCop = F1opH * I1opH * Frop * Iop * Fop
    if transpose:
        MDCop = TopH * MDCop * Top

    # force dtype to be real (as FFT operators assume real inputs and outputs)
    MDCop.dtype = rdtype

    return MDCop


def MDC(G, nt, nv, dt=1., dr=1., twosided=True, fast=None,
        dtype=None, fftengine='numpy', transpose=True,
        saveGt=True, conj=False, usematmul=False, prescaled=False):
    r"""Multi-dimensional convolution.

    Apply multi-dimensional convolution between two datasets. If
    ``transpose=True``, model and data should be provided after flattening
    2- or 3-dimensional arrays of size :math:`[n_r (\times n_{vs}) \times n_t]`
    and :math:`[n_s (\times n_{vs}) \times n_t]` (or :math:`2*n_t-1` for
    ``twosided=True``), respectively. If ``transpose=False``, model and data
    should be provided after flattening 2- or 3-dimensional arrays of size
    :math:`[n_t \times n_r (\times n_{vs})]` and
    :math:`[n_t \times n_s (\times n_{vs})]` (or :math:`2*n_t-1` for
    ``twosided=True``), respectively.

    .. warning:: A new implementation of MDC is provided in v1.5.0. This
      currently affects only the inner working of the operator and end-users
      can use the operator in the same way as they used to do with the previous
      one. Nevertheless, it is now reccomended to use the operator with
      ``transpose=False``, as this behaviour will become default in version
      v2.0.0 and the behaviour with ``transpose=True`` will be deprecated.

    Parameters
    ----------
    G : :obj:`numpy.ndarray`
        Multi-dimensional convolution kernel in frequency domain of size
        :math:`[n_s \times n_r \times n_{fmax}]` if ``transpose=True``
        or size :math:`[n_{fmax} \times n_s \times n_r]` if ``transpose=False``
    nt : :obj:`int`
        Number of samples along time axis
    nv : :obj:`int`
        Number of samples along virtual source axis
    dt : :obj:`float`, optional
        Sampling of time integration axis
    dr : :obj:`float`, optional
        Sampling of receiver integration axis
    twosided : :obj:`bool`, optional
        MDC operator has both negative and positive time (``True``) or
        only positive (``False``)
    fast : :obj:`bool`, optional
        *Deprecated*, will be removed in v2.0.0
    dtype : :obj:`str`, optional
        *Deprecated*, will be removed in v2.0.0
    fftengine : :obj:`str`, optional
        Engine used for fft computation (``numpy`` or ``fftw``)
    transpose : :obj:`bool`, optional
        Transpose ``G`` and inputs such that time/frequency is placed in first
        dimension. This allows back-compatibility with v1.4.0 and older but
        will be removed in v2.0.0 where time/frequency axis will be required
        to be in first dimension for efficiency reasons.
    saveGt : :obj:`bool`, optional
        Save ``G`` and ``G^H`` to speed up the computation of adjoint of
        :class:`pylops.signalprocessing.Fredholm1` (``True``) or create
        ``G^H`` on-the-fly (``False``) Note that ``saveGt=True`` will be
        faster but double the amount of required memory
    conj : :obj:`str`, optional
        Perform Fredholm integral computation with complex conjugate of ``G``
    usematmul : :obj:`bool`, optional
        Use :func:`numpy.matmul` (``True``) or for-loop with :func:`numpy.dot`
        (``False``) in :py:class:`pylops.signalprocessing.Fredholm1` operator.
        Refer to Fredholm1 documentation for details.
    prescaled : :obj:`bool`, optional
        Apply scaling to kernel (``False``) or not (``False``) when performing
        spatial and temporal summations. In case ``prescaled=True``, the
        kernel is assumed to have been pre-scaled when passed to the MDC
        routine.

    Raises
    ------
    ValueError
        If ``nt`` is even and ``twosided=True``

    See Also
    --------
    MDD : Multi-dimensional deconvolution

    Notes
    -----
    The so-called multi-dimensional convolution (MDC) is a chained
    operator [1]_. It is composed of a forward Fourier transform,
    a multi-dimensional integration, and an inverse Fourier transform:

    .. math::
        y(t, s, v) = \mathscr{F}^{-1} \Big( \int_S G(f, s, r)
        \mathscr{F}(x(t, r, v)) dr \Big)

    which is discretized as follows:

    .. math::
        y(t, s, v) = \mathscr{F}^{-1} \Big( \sum_{i_r=0}^{n_r}
        (\sqrt{n_t} * d_t * d_r) G(f, s, i_r) \mathscr{F}(x(t, i_r, v)) \Big)

    where :math:`(\sqrt{n_t} * d_t * d_r)` is not applied if ``prescaled=True``.

    This operation can be discretized and performed by means of a
    linear operator

    .. math::
        \mathbf{D}= \mathbf{F}^H  \mathbf{G} \mathbf{F}

    where :math:`\mathbf{F}` is the Fourier transform applied along
    the time axis and :math:`\mathbf{G}` is the multi-dimensional
    convolution kernel.

    .. [1] Wapenaar, K., van der Neut, J., Ruigrok, E., Draganov, D., Hunziker,
       J., Slob, E., Thorbecke, J., and Snieder, R., "Seismic interferometry
       by crosscorrelation and by multi-dimensional deconvolution: a
       systematic comparison", Geophysical Journal International, vol. 185,
       pp. 1335-1364. 2011.

    """
    return _MDC(G, nt, nv, dt=dt, dr=dr, twosided=twosided, fast=fast,
                dtype=dtype, transpose=transpose, saveGt=saveGt,
                conj=conj, prescaled=prescaled,
                args_FFT={'engine': fftengine},
                args_Fredholm1={'usematmul': usematmul})


def MDD(G, d, dt=0.004, dr=1., nfmax=None, wav=None,
        twosided=True, causality_precond=False, adjoint=False,
        psf=False, dtype='float64', dottest=False,
        saveGt=True, add_negative=True, smooth_precond=0, **kwargs_lsqr):
    r"""Multi-dimensional deconvolution.

    Solve multi-dimensional deconvolution problem using
    :py:func:`scipy.sparse.linalg.lsqr` iterative solver.

    Parameters
    ----------
    G : :obj:`numpy.ndarray`
        Multi-dimensional convolution kernel in time domain of size
        :math:`[n_s \times n_r \times n_t]` for ``twosided=False`` or
        ``twosided=True`` and ``add_negative=True``
        (with only positive times) or size
        :math:`[n_s \times n_r \times 2*n_t-1]` for ``twosided=True`` and
        ``add_negative=False``
        (with both positive and negative times)
    d : :obj:`numpy.ndarray`
        Data in time domain :math:`[n_s (\times n_vs) \times n_t]` if
        ``twosided=False`` or ``twosided=True`` and ``add_negative=True``
        (with only positive times) or size
        :math:`[n_s (\times n_vs) \times 2*n_t-1]` if ``twosided=True``
    dt : :obj:`float`, optional
        Sampling of time integration axis
    dr : :obj:`float`, optional
        Sampling of receiver integration axis
    nfmax : :obj:`int`, optional
        Index of max frequency to include in deconvolution process
    wav : :obj:`numpy.ndarray`, optional
        Wavelet to convolve to the inverted model and psf
        (must be centered around its index in the middle of the array).
        If ``None``, the outputs of the inversion are returned directly.
    twosided : :obj:`bool`, optional
        MDC operator and data both negative and positive time (``True``)
        or only positive (``False``)
    add_negative : :obj:`bool`, optional
        Add negative side to MDC operator and data (``True``) or not
        (``False``)- operator and data are already provided with both positive
        and negative sides. To be used only with ``twosided=True``.
    causality_precond : :obj:`bool`, optional
        Apply causality mask (``True``) or not (``False``)
    adjoint : :obj:`bool`, optional
        Compute and return adjoint(s)
    psf : :obj:`bool`, optional
        Compute and return Point Spread Function (PSF) and its inverse
    dtype : :obj:`bool`, optional
        Type of elements in input array.
    dottest : :obj:`bool`, optional
        Apply dot-test
    saveGt : :obj:`bool`, optional
        Save ``G`` and ``G^H`` to speed up the computation of adjoint of
        :class:`pylops.signalprocessing.Fredholm1` (``True``) or create
        ``G^H`` on-the-fly (``False``) Note that ``saveGt=True`` will be
        faster but double the amount of required memory
    **kwargs_lsqr
        Arbitrary keyword arguments for
        :py:func:`scipy.sparse.linalg.lsqr` solver

    Returns
    -------
    minv : :obj:`numpy.ndarray`
        Inverted model of size :math:`[n_r (\times n_{vs}) \times n_t]`
        for ``twosided=False`` or
        :math:`[n_r (\times n_vs) \times 2*n_t-1]` for ``twosided=True``
    madj : :obj:`numpy.ndarray`
        Adjoint model of size :math:`[n_r (\times n_{vs}) \times n_t]`
        for ``twosided=False`` or
        :math:`[n_r (\times n_r) \times 2*n_t-1]` for ``twosided=True``
    psfinv : :obj:`numpy.ndarray`
        Inverted psf of size :math:`[n_r \times n_r \times n_t]`
        for ``twosided=False`` or
        :math:`[n_r \times n_r \times 2*n_t-1]` for ``twosided=True``
    psfadj : :obj:`numpy.ndarray`
        Adjoint psf of size :math:`[n_r \times n_r \times n_t]`
        for ``twosided=False`` or
        :math:`[n_r \times n_r \times 2*n_t-1]` for ``twosided=True``

    See Also
    --------
    MDC : Multi-dimensional convolution

    Notes
    -----
    Multi-dimensional deconvolution (MDD) is a mathematical ill-solved problem,
    well-known in the image processing and geophysical community [1]_.

    MDD aims at removing the effects of a Multi-dimensional Convolution
    (MDC) kernel or the so-called blurring operator or point-spread
    function (PSF) from a given data. It can be written as

    .. math::
        \mathbf{d}= \mathbf{D} \mathbf{m}

    or, equivalently, by means of its normal equation

    .. math::
        \mathbf{m}= (\mathbf{D}^H\mathbf{D})^{-1} \mathbf{D}^H\mathbf{d}

    where :math:`\mathbf{D}^H\mathbf{D}` is the PSF.

    .. [1] Wapenaar, K., van der Neut, J., Ruigrok, E., Draganov, D., Hunziker,
       J., Slob, E., Thorbecke, J., and Snieder, R., "Seismic interferometry
       by crosscorrelation and by multi-dimensional deconvolution: a
       systematic comparison", Geophyscial Journal International, vol. 185,
       pp. 1335-1364. 2011.

    """
    ns, nr, nt = G.shape
    if len(d.shape) == 2:
        nv = 1
    else:
        nv = d.shape[1]
    if twosided:
        if add_negative:
            nt2 = 2 * nt - 1
        else:
            nt2 = nt
            nt = (nt2 + 1) // 2
        nfmax_allowed = int(np.ceil((nt2+1)/2))
    else:
        nt2 = nt
        nfmax_allowed = nt

    # Fix nfmax to be at maximum equal to half of the size of fft samples
    if nfmax is None or nfmax > nfmax_allowed:
        nfmax = nfmax_allowed
        logging.warning('nfmax set equal to ceil[(nt+1)/2=%d]' % nfmax)

    # Add negative part to data and model
    if twosided and add_negative:
        G = np.concatenate((np.zeros((ns, nr, nt - 1)), G), axis=-1)
        d = np.concatenate((np.squeeze(np.zeros((ns, nv, nt - 1))), d),
                           axis=-1)

    # Bring kernel to frequency domain
    Gfft = np.fft.rfft(G, nt2, axis=-1)
    Gfft = Gfft[..., :nfmax]

    # Bring frequency/time to first dimension
    Gfft = np.moveaxis(Gfft, -1, 0)
    d = np.moveaxis(d, -1, 0)
    if psf:
        G = np.moveaxis(G, -1, 0)

    # Define MDC linear operator
    MDCop = MDC(Gfft, nt2, nv=nv, dt=dt, dr=dr, twosided=twosided,
                transpose=False, saveGt=saveGt)
    if psf:
        PSFop = MDC(Gfft, nt2, nv=nr, dt=dt, dr=dr, twosided=twosided,
                    transpose=False, saveGt=saveGt)
    if dottest:
        Dottest(MDCop, nt2*ns*nv, nt2*nr*nv, verb=True)
        if psf:
            Dottest(PSFop, nt2 * ns * nr, nt2 * nr * nr, verb=True)

    # Adjoint
    if adjoint:
        madj = MDCop.H * d.flatten()
        madj = np.squeeze(madj.reshape(nt2, nr, nv))
        madj = np.moveaxis(madj, 0, -1)
        if psf:
            psfadj = PSFop.H * G.flatten()
            psfadj = np.squeeze(psfadj.reshape(nt2, nr, nr))
            psfadj = np.moveaxis(psfadj, 0, -1)

    # Inverse
    if twosided and causality_precond:
        P = np.ones((nt2, nr, nv))
        P[:nt - 1] = 0
        if smooth_precond > 0:
            P = filtfilt(np.ones(smooth_precond)/smooth_precond, 1, P, axis=0)
        Pop = Diagonal(P)
        minv = PreconditionedInversion(MDCop, Pop, d.flatten(),
                                       returninfo=False, **kwargs_lsqr)
    else:
        minv = lsqr(MDCop, d.flatten(), **kwargs_lsqr)[0]
    minv = np.squeeze(minv.reshape(nt2, nr, nv))
    minv = np.moveaxis(minv, 0, -1)
    if wav is not None:
        minv = sp_convolve1d(minv, wav, axis=-1)

    if psf:
        psfinv = lsqr(PSFop, G.flatten(), **kwargs_lsqr)[0]
        psfinv = np.squeeze(psfinv.reshape(nt2, nr, nr))
        psfinv = np.moveaxis(psfinv, 0, -1)
        if wav is not None:
            psfinv = sp_convolve1d(psfinv, wav, axis=-1)

    if adjoint and psf:
        return minv, madj, psfinv, psfadj
    elif adjoint:
        return minv, madj
    elif psf:
        return minv, psfinv
    else:
        return minv



import logging
import numpy as np

from pylops import LinearOperator
from ._ChirpRadon3D import _chirp_radon_3d, _chirp_radon_3d_fftw

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


class ChirpRadon3D(LinearOperator):
    r"""3D Chirp Radon transform

    Apply Radon forward (and adjoint) transform using Fast Fourier Transform
    and Chirp functions to a 3-dimensional array of size
    :math:`[n_x \times n_t]` (and :math:`[n_{x} \times n_t]`).

    Note that forward and adjoint are swapped compared to the time-space
    implementation in :class:`pylops.signalprocessing.Radon3D` and a direct
    `inverse` method is also available for this implementation.

    Parameters
    ----------
    taxis : :obj:`np.ndarray`
        Time axis
    hxaxis : :obj:`np.ndarray`
        Fast patial axis
    hyaxis : :obj:`np.ndarray`
        Slow spatial axis
    pmax : :obj:`np.ndarray`
        Two element array :math:`(p_{y,max}, p_{x,max})` of :math:`\tan`
        of maximum stacking angles in :math:`y` and :math:`x` directions
        :math:`(\tan(\alpha_{y,max}), \tan(\alpha_{x,max}))`. If one operates
        in terms of minimum velocity :math:`c_0`, then
        :math:`p_{y.max}=c_0dy/dt` and :math:`p_{x,max}=c_0dx/dt`
    engine : :obj:`str`, optional
        Engine used for fft computation (``numpy`` or ``fftw``)
    dtype : :obj:`str`, optional
        Type of elements in input array.
    **kwargs_fftw
            Arbitrary keyword arguments for :py:class:`pyfftw.FTTW`
            (reccomended: ``flags=('FFTW_ESTIMATE', ), threads=NTHREADS``)

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    Refer to [1]_ for the theoretical and implementation details.

    .. [1] Andersson, F and Robertsson J. "Fast :math:`\tau-p` transforms by
        chirp modulation", Geophysics, vol 84, NO.1, pp. A13-A17, 2019.

    """
    def __init__(self, taxis, hyaxis, hxaxis, pmax, engine='numpy',
                 dtype='float64', **kwargs_fftw):
        self.dt = taxis[1] - taxis[0]
        self.dy = hyaxis[1] - hyaxis[0]
        self.dx = hxaxis[1] - hxaxis[0]
        self.nt, self.nx, self.ny = taxis.size, hxaxis.size, hyaxis.size
        self.pmax = pmax
        if engine == 'numpy':
            self.func = lambda x, mode: _chirp_radon_3d(x, self.dt,
                                                        self.dy, self.dx,
                                                        self.pmax, mode=mode)
        elif engine == 'fftw':
            self.func = lambda x, mode: _chirp_radon_3d_fftw(x, self.dt,
                                                             self.dy, self.dx,
                                                             self.pmax,
                                                             mode=mode,
                                                             **kwargs_fftw)
        else:
            raise NotImplementedError('engine must be numpy or fftw')

        self.shape = (self.nt * self.nx * self.ny,
                      self.nt * self.nx * self.ny)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        x = x.reshape(self.ny, self.nx, self.nt)
        y = self.func(x, mode='f', )
        return y.ravel()

    def _rmatvec(self, x):
        x = x.reshape(self.ny, self.nx, self.nt)
        y = self.func(x, mode='a')
        return y.ravel()

    def inverse(self, x):
        x = x.reshape(self.ny, self.nx, self.nt)
        y = self.func(x, mode='i')
        return y.ravel()



import logging

import numpy as np
from pylops import LinearOperator

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)

class Regression(LinearOperator):
    r"""Polynomial regression.

    Creates an operator that applies polynomial regression to a set of points.
    Values along the t-axis must be provided while initializing the operator.
    The coefficients of the polynomial regression form the model vector to
    be provided in forward mode, while the values of the regression
    curve shall be provided in adjoint mode.

    Parameters
    ----------
    taxis : :obj:`numpy.ndarray`
        Elements along the t-axis.
    order : :obj:`int`
        Order of the regressed polynomial.
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)

    Raises
    ------
    TypeError
        If ``t`` is not :obj:`numpy.ndarray`.

    See Also
    --------
    LinearRegression: Linear regression

    Notes
    -----
    The Regression operator solves the following problem:

    .. math::
        y_i = \sum_{n=0}^{order} x_n t_i^n  \qquad \forall i=1,2,...,N

    where :math:`N` represents the order of the chosen polynomial. We can
    express this problem in a matrix form

    .. math::
        \mathbf{y}=  \mathbf{A} \mathbf{x}

    where

    .. math::
        \mathbf{y}= [y_1, y_2,...,y_N]^T,
        \qquad \mathbf{x}= [x_0, x_1,...,x_{order}]^T

    and

    .. math::
        \mathbf{A}
        = \begin{bmatrix}
            1  & t_{1} & t_{1}^2 & .. & t_{1}^{order}  \\
            1  & t_{2} & t_{2}^2 & .. & t_{1}^{order}  \\
            .. & ..    & ..      & .. & ..             \\
            1  & t_{N} & t_{N}^2 & .. & t_{N}^{order}  \\
        \end{bmatrix}

    """
    def __init__(self, taxis, order, dtype='float64'):
        if not isinstance(taxis, np.ndarray):
            logging.error('t must be numpy.ndarray...')
            raise TypeError('t must be numpy.ndarray...')
        else:
            self.taxis = taxis
        self.order = order
        self.shape = (len(self.taxis), self.order+1)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        y = np.zeros_like(self.taxis)
        for i in range(self.order+1):
            y += x[i]*self.taxis**i
        return y

    def _rmatvec(self, x):
        return np.vstack([np.dot(self.taxis**i, x)
                          for i in range(self.order+1)])

    def apply(self, t, x):
        """Return values along y-axis given certain ``t`` location(s) along
        t-axis and regression coefficients ``x``

        Parameters
        ----------
        taxis : :obj:`numpy.ndarray`
            Elements along the t-axis.
        x : :obj:`numpy.ndarray`
            Regression coefficients
        dtype : :obj:`str`, optional
        Returns
        ----------
        y : :obj:`numpy.ndarray`
            Values along y-axis

        """
        torig = self.taxis.copy()
        self.taxis = t
        y = self._matvec(x)
        self.taxis = torig
        return y



from __future__ import division

import logging
import numpy as np
from scipy.linalg import solve, lstsq
from scipy.sparse.linalg import LinearOperator as spLinearOperator
from scipy.sparse.linalg import spsolve, lsqr
from scipy.linalg import eigvals
from scipy.sparse.linalg import eigs as sp_eigs
from scipy.sparse.linalg import eigsh as sp_eigsh
from scipy.sparse.linalg import lobpcg as sp_lobpcg
from scipy.sparse import csr_matrix

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


class LinearOperator(spLinearOperator):
    """Common interface for performing matrix-vector products.

    This class is an overload of the
    :py:class:`scipy.sparse.linalg.LinearOperator` class. It adds
    functionalities by overloading standard operators such as ``__truediv__``
    as well as creating convenience methods such as ``eigs``, ``cond``, and
    ``conj``.

    .. note:: End users of PyLops should not use this class directly but simply
      use operators that are already implemented. This class is meant for
      developers and it has to be used as the parent class of any new operator
      developed within PyLops. Find more details regarding implementation of
      new operators at :ref:`addingoperator`.

    Parameters
    ----------
    Op : :obj:`scipy.sparse.linalg.LinearOperator` or :obj:`scipy.sparse.linalg._ProductLinearOperator` or :obj:`scipy.sparse.linalg._SumLinearOperator`
        Operator
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (``True``) or not (``False``)

    """
    def __init__(self, Op=None, explicit=False):
        self.explicit = explicit
        if Op is not None:
            self.Op = Op
            self.shape = self.Op.shape
            self.dtype = self.Op.dtype

    def _matvec(self, x):
        if callable(self.Op._matvec):
            return self.Op._matvec(x)

    def _rmatvec(self, x):
        if callable(self.Op._rmatvec):
            return self.Op._rmatvec(x)

    def _matmat(self, X):
        """Matrix-matrix multiplication handler.

        Modified version of scipy _matmat to avoid having trailing dimension
        in col when provided to matvec
        """
        return np.vstack([self.matvec(col.reshape(-1)) for col in X.T]).T

    def __mul__(self, x):
        y = super().__mul__(x)
        if isinstance(y, spLinearOperator):
            y = aslinearoperator(y)
        return y

    def __rmul__(self, x):
        return aslinearoperator(super().__rmul__(x))

    def __pow__(self, p):
        return aslinearoperator(super().__pow__(p))

    def __add__(self, x):
        return aslinearoperator(super().__add__(x))

    def __neg__(self):
        return aslinearoperator(super().__neg__())

    def __sub__(self, x):
        return aslinearoperator(super().__sub__(x))

    def _adjoint(self):
        return aslinearoperator(super()._adjoint())

    def div(self, y, niter=100):
        r"""Solve the linear problem :math:`\mathbf{y}=\mathbf{A}\mathbf{x}`.

        Overloading of operator ``/`` to improve expressivity of `Pylops`
        when solving inverse problems.

        Parameters
        ----------
        y : :obj:`np.ndarray`
            Data
        niter : :obj:`int`, optional
            Number of iterations (to be used only when ``explicit=False``)

        Returns
        -------
        xest : :obj:`np.ndarray`
            Estimated model

        """
        xest = self.__truediv__(y, niter=niter)
        return xest

    def __truediv__(self, y, niter=100):
        if self.explicit is True:
            if isinstance(self.A, np.ndarray):
                if self.A.shape[0] == self.A.shape[1]:
                    xest = solve(self.A, y)
                else:
                    xest = lstsq(self.A, y)[0]
            else:
                xest = spsolve(self.A, y)
        else:
            xest = lsqr(self, y, iter_lim=niter)[0]
        return xest

    def todense(self):
        r"""Return dense matrix.

        The operator in converted into its dense matrix equivalent. In order
        to do so, the operator is applied to an identity matrix whose number
        of rows and columns is equivalent to the number of columns of the
        operator. Note that this operation may be costly for very large
        operators and it is only suggest it to use as a way to inspect the
        structure of the matricial equivalent of the operator.

        Returns
        -------
        matrix : :obj:`numpy.ndarray`
            Dense matrix.

        """
        # Wrap self into a LinearOperator. This is done for cases where self
        # is a _SumLinearOperator or _ProductLinearOperator, so that it regains
        # the dense method
        Op = aslinearoperator(self)

        identity = np.eye(self.shape[1], dtype=self.dtype)
        matrix = Op.matmat(identity)
        return matrix


    def tosparse(self):
        r"""Return sparse matrix.

        The operator in converted into its sparse (CSR) matrix equivalent. In order
        to do so, the operator is applied to series of unit vectors with length equal
        to the number of coloumns in the original operator.

        Returns
        -------
        matrix : :obj:`scipy.sparse.csr_matrix`
            Sparse matrix.

        """

        Op = aslinearoperator(self)
        (_, n) = self.shape

        # stores non-zero data for the sparse matrix creation
        entries = []
        indices = []

        # loop through columns of self
        for i in range(n):

            # make a unit vector for the ith column
            unit_i = np.zeros(n)
            unit_i[i] = 1

            # multiply unit vector to self and find the non-zeros
            res_i = Op * unit_i
            rows_nz = np.where(res_i != 0)[0]

            # append the non-zero values and indices to the lists
            for j in rows_nz:
                indices.append([i, j])
            entries_i = res_i[rows_nz]
            for e in entries_i:
                entries.append(e)

        # post process the entries / indices for scipy.sparse.csr_matrix
        entries = np.array(entries)
        indices = np.array(indices)
        i, j = indices[:,0], indices[:,1]

        # construct a sparse, CSR matrix from the entries / indices data.
        matrix = csr_matrix((entries, (j, i)), shape=self.shape, dtype=self.dtype)
        return matrix

    def eigs(self, neigs=None, symmetric=False, niter=None,
             uselobpcg=False, **kwargs_eig):
        r"""Most significant eigenvalues of linear operator.

        Return an estimate of the most significant eigenvalues
        of the linear operator. If the operator has rectangular
        shape (``shape[0]!=shape[1]``), eigenvalues are first
        computed for the square operator :math:`\mathbf{A^H}\mathbf{A}`
        and the square-root values are returned.

        Parameters
        ----------
        neigs : :obj:`int`
            Number of eigenvalues to compute (if ``None``, return all). Note
            that for ``explicit=False``, only :math:`N-1` eigenvalues can be
            computed where :math:`N` is the size of the operator in the
            model space
        symmetric : :obj:`bool`, optional
            Operator is symmetric (``True``) or not (``False``). User should
            set this parameter to ``True`` only when it is guaranteed that the
            operator is real-symmetric or complex-hermitian matrices
        niter : :obj:`int`, optional
            Number of iterations for eigenvalue estimation
        uselobpcg : :obj:`bool`, optional
            Use :func:`scipy.sparse.linalg.lobpcg`
        **kwargs_eig
            Arbitrary keyword arguments for :func:`scipy.sparse.linalg.eigs`,
            :func:`scipy.sparse.linalg.eigsh`, or
            :func:`scipy.sparse.linalg.lobpcg`

        Returns
        -------
        eigenvalues : :obj:`numpy.ndarray`
            Operator eigenvalues.

        Raises
        -------
        ValueError
            If ``uselobpcg=True`` for a non-symmetric square matrix with
            complex type

        Notes
        -----
        Depending on the size of the operator, whether it is explicit or not
        and the number of eigenvalues requested, different algorithms are
        used by this routine.

        More precisely, when only a limited number of eigenvalues is requested
        the :func:`scipy.sparse.linalg.eigsh` method is used in case of
        ``symmetric=True`` and the :func:`scipy.sparse.linalg.eigs` method
        is used ``symmetric=False``. However, when the matrix is represented
        explicitly within the linear operator (``explicit=True``) and all the
        eigenvalues are requested the :func:`scipy.linalg.eigvals` is used
        instead.

        Finally, when only a limited number of eigenvalues is required,
        it is also possible to explicitly choose to use the
        ``scipy.sparse.linalg.lobpcg`` method via the ``uselobpcg`` input
        parameter flag.

        Most of these algorithms are a port of ARPACK [1]_, a Fortran package
        which provides routines for quickly finding eigenvalues/eigenvectors
        of a matrix. As ARPACK requires only left-multiplication
        by the matrix in question, eigenvalues/eigenvectors can also be
        estimated for linear operators when the dense matrix is not available.

        .. [1] http://www.caam.rice.edu/software/ARPACK/

        """
        if self.explicit and isinstance(self.A, np.ndarray):
            if self.shape[0] == self.shape[1]:
                if neigs is None or neigs == self.shape[1]:
                    eigenvalues = eigvals(self.A)
                else:
                    if not symmetric and np.iscomplexobj(self) and uselobpcg:
                        raise ValueError('cannot use scipy.sparse.linalg.lobpcg '
                                         'for non-symmetric square matrices of '
                                         'complex type...')
                    if symmetric and uselobpcg:
                        X = np.random.rand(self.shape[0], neigs)
                        eigenvalues = \
                            sp_lobpcg(self.A, X=X, maxiter=niter,
                                      **kwargs_eig)[0]
                    elif symmetric:
                        eigenvalues = sp_eigsh(self.A, k=neigs, maxiter=niter,
                                               **kwargs_eig)[0]
                    else:
                        eigenvalues = sp_eigs(self.A, k=neigs, maxiter=niter,
                                              **kwargs_eig)[0]

            else:
                if neigs is None or neigs == self.shape[1]:
                    eigenvalues = np.sqrt(eigvals(np.dot(np.conj(self.A.T),
                                                         self.A)))
                else:
                    if uselobpcg:
                        X = np.random.rand(self.shape[1], neigs)
                        eigenvalues = np.sqrt(sp_lobpcg(
                            np.dot(np.conj(self.A.T), self.A),
                            X=X, maxiter=niter, **kwargs_eig)[0])
                    else:
                        eigenvalues = np.sqrt(sp_eigsh(
                            np.dot(np.conj(self.A.T), self.A),
                            k=neigs, maxiter=niter, **kwargs_eig)[0])
        else:
            if neigs is None or neigs >= self.shape[1]:
                neigs = self.shape[1] - 2
            if self.shape[0] == self.shape[1]:
                if not symmetric and np.iscomplexobj(self) and uselobpcg:
                    raise ValueError('cannot use scipy.sparse.linalg.lobpcg for '
                                     'non symmetric square matrices of '
                                     'complex type...')
                if symmetric and uselobpcg:
                    X = np.random.rand(self.shape[0], neigs)
                    eigenvalues = \
                        sp_lobpcg(self, X=X, maxiter=niter, **kwargs_eig)[0]
                elif symmetric:
                    eigenvalues = sp_eigsh(self, k=neigs, maxiter=niter,
                                           **kwargs_eig)[0]
                else:
                    eigenvalues = sp_eigs(self, k=neigs, maxiter=niter,
                                          **kwargs_eig)[0]
            else:
                if uselobpcg:
                    X = np.random.rand(self.shape[1], neigs)
                    eigenvalues = np.sqrt(sp_lobpcg(self.H * self, X=X,
                                                    maxiter=niter,
                                                    **kwargs_eig)[0])
                else:
                    eigenvalues = np.sqrt(sp_eigs(self.H * self, k=neigs,
                                                  maxiter=niter,
                                                  **kwargs_eig)[0])
        return -np.sort(-eigenvalues)

    def cond(self, uselobpcg=False, **kwargs_eig):
        r"""Condition number of linear operator.

        Return an estimate of the condition number of the linear operator as
        the ratio of the largest and lowest estimated eigenvalues.

        Parameters
        ----------
        uselobpcg : :obj:`bool`, optional
            Use :func:`scipy.sparse.linalg.lobpcg` to compute eigenvalues
        **kwargs_eig
            Arbitrary keyword arguments for :func:`scipy.sparse.linalg.eigs`,
            :func:`scipy.sparse.linalg.eigsh`, or
            :func:`scipy.sparse.linalg.lobpcg`

        Returns
        -------
        eigenvalues : :obj:`numpy.ndarray`
            Operator eigenvalues.

        Notes
        -----
        The condition number of a matrix (or linear operator) can be estimated
        as the ratio of the largest and lowest estimated eigenvalues:

        .. math::
            k= \frac{\lambda_{max}}{\lambda_{min}}

        The condition number provides an indication of the rate at which the
        solution of the inversion of the linear operator :math:`A` will
        change with respect to a change in the data :math:`y`.

        Thus, if the condition number is large, even a small error in :math:`y`
        may cause a large error in :math:`x`. On the other hand, if the
        condition number is small then the error in :math:`x` is not much
        bigger than the error in :math:`y`. A problem with a low condition
        number is said to be *well-conditioned*, while a problem with a high
        condition number is said to be *ill-conditioned*.

        """
        if not uselobpcg:
            cond = np.asscalar(self.eigs(neigs=1, which='LM', **kwargs_eig))/ \
                   np.asscalar(self.eigs(neigs=1, which='SM', **kwargs_eig))
        else:
            print('here')
            cond = np.asscalar(self.eigs(neigs=1, uselobpcg=True, largest=True,
                                         **kwargs_eig)) / \
                   np.asscalar(self.eigs(neigs=1, uselobpcg=True, largest=False,
                                         **kwargs_eig))

        return cond

    def conj(self):
        """Complex conjugate operator

        Returns
        -------
        conjop : :obj:`pylops.LinearOperator`
            Complex conjugate operator

        """
        conjop = _ConjLinearOperator(self)
        return conjop

    def apply_columns(self, cols):
        """Apply subset of columns of operator

        This method can be used to wrap a LinearOperator and mimic the action
        of a subset of columns of the operator on a reduced model in forward
        mode, and retrieve only the result of a subset of rows in adjoint mode.

        Note that unless the operator has ``explicit=True``, this is not
        optimal as the entire forward and adjoint passes of the original
        operator will have to be perfomed. It can however be useful for the
        implementation of solvers such as Orthogonal Matching Pursuit (OMP)
        that iteratively build a solution by evaluate only a subset of the
        columns of the operator.

        Parameters
        ----------
        cols : :obj:`list`
            Columns to be selected

        Returns
        -------
        colop : :obj:`pylops.LinearOperator`
            Apply column operator

        """
        colop = _ColumnLinearOperator(self, cols)
        return colop


class _ConjLinearOperator(LinearOperator):
    """Complex conjugate linear operator
    """
    def __init__(self, Op):
        if not isinstance(Op, spLinearOperator):
            raise TypeError('Op must be a LinearOperator')
        super(_ConjLinearOperator, self).__init__(Op, Op.shape)
        self.Op = Op

    def _matvec(self, x):
        return (self.Op._matvec(x.conj())).conj()

    def _rmatvec(self, x):
        return (self.Op._rmatvec(x.conj())).conj()

    def _adjoint(self):
        return _ConjLinearOperator(self.Op.H)


class _ColumnLinearOperator(LinearOperator):
    """Column selector linear operator

    Produces the forward and adjoint passes with a subset of columns of the
    original operator
    """
    def __init__(self, Op, cols):
        if not isinstance(Op, spLinearOperator):
            raise TypeError('Op must be a LinearOperator')
        super(_ColumnLinearOperator, self).__init__(Op, Op.explicit)
        self.Op = Op
        self.cols = cols
        self.shape = (Op.shape[0], len(cols))
        if self.explicit:
            self.Opcol = Op.A[:, cols]

    def _matvec(self, x):
        if self.explicit:
            y = self.Opcol @ x
        else:
            y = np.zeros(self.Op.shape[1], dtype=self.dtype)
            y[self.cols] = x
            y = self.Op._matvec(y)
        return y

    def _rmatvec(self, x):
        if self.explicit:
            y = self.Opcol.T @ x
        else:
            y = self.Op._rmatvec(x)
            y = y[self.cols]
        return y


def aslinearoperator(Op):
    """Return Op as a LinearOperator.

    Converts any operator into a LinearOperator. This can be used when `Op`
    is a private operator to ensure that the return operator has all properties
    and methods of the parent class.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator` or any other Operator
        Operator of any type

    Returns
    -------
    Op : :obj:`pylops.LinearOperator`
        Operator of type :obj:`pylops.LinearOperator`

    """
    if isinstance(Op, LinearOperator):
        return Op
    else:
        return LinearOperator(Op)


import numpy as np
from pylops.signalprocessing import Convolve2D


def Smoothing2D(nsmooth, dims, nodir=None, dtype='float64'):
    r"""2D Smoothing.

    Apply smoothing to model (and data) along two directions of a
    multi-dimensional array depending on the choice of ``nodir``.

    Parameters
    ----------
    nsmooth : :obj:`tuple` or :obj:`list`
        Lenght of smoothing operatorin 1st and 2nd dimensions (must be odd)
    dims : :obj:`tuple`
        Number of samples for each dimension
    nodir : :obj:`int`, optional
        Direction along which smoothing is NOT applied (set to None for 2d
        arrays)
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

     See Also
    --------
    lops.signalprocessing.Convolve2D : 2D convolution

    Notes
    -----
    The 2D Smoothing operator is a special type of convolutional operator that
    convolves the input model (or data) with a constant 2d filter of size
    :math:`n_{smooth, 1} \quad x \quad n_{smooth, 2}`:

    Its application to a two dimensional input signal is:

    .. math::
        y[i,j] = 1/(n_{smooth, 1}*n_{smooth, 2})
        \sum_{l=-(n_{smooth,1}-1)/2}^{(n_{smooth,1}-1)/2}
        \sum_{m=-(n_{smooth,2}-1)/2}^{(n_{smooth,2}-1)/2} x[l,m]

    Note that since the filter is symmetrical, the *Smoothing2D* operator is
    self-adjoint.

    """
    if isinstance(nsmooth, tuple):
        nsmooth = list(nsmooth)
    if nsmooth[0] % 2 == 0:
        nsmooth[0] += 1
    if nsmooth[1] % 2 == 0:
        nsmooth[1] += 1

    h = np.ones((nsmooth[0], nsmooth[1]))/float(nsmooth[0]*nsmooth[1])
    return Convolve2D(np.prod(np.array(dims)), h=h,
                      offset=[(nsmooth[0]-1)/2, (nsmooth[1]-1)/2],
                      dims=dims, nodir=nodir, dtype=dtype)



import logging
import numpy as np
from pylops import LinearOperator
from pylops.basicoperators import Restriction, Diagonal, MatrixMult, Transpose

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)

def _checkunique(iava):
    _, count = np.unique(iava, return_counts=True)
    if np.any(count > 1):
        raise ValueError('Repeated values in iava array')

def _nearestinterp(M, iava, dims=None, dir=0, dtype='float64'):
    """Nearest neighbour interpolation.
    """
    iava = np.round(iava).astype(np.int)
    _checkunique(iava)
    return Restriction(M, iava, dims=dims, dir=dir, dtype=dtype), iava

def _linearinterp(M, iava, dims=None, dir=0, dtype='float64'):
    """Linear interpolation.
    """
    if np.issubdtype(iava.dtype, np.integer):
        iava = iava.astype(np.float)
    if dims is None:
        lastsample = M
        dimsd = None
    else:
        lastsample = dims[dir]
        dimsd = list(dims)
        dimsd[dir] = len(iava)
        dimsd = tuple(dimsd)

    # ensure that samples are not beyond the last sample, in that case set to
    # penultimate sample and raise a warning
    outside = (iava >= lastsample - 1)
    if sum(outside) > 0:
        logging.warning('at least one value is beyond penultimate sample, '
                        'forced to be at penultimate sample')
    iava[outside] = lastsample - 1 - 1e-10
    _checkunique(iava)

    # find indices and weights
    iva_l = np.floor(iava).astype(np.int)
    iva_r = iva_l + 1
    weights = iava - iva_l

    # create operators
    Op = Diagonal(1 - weights, dims=dimsd, dir=dir, dtype=dtype) * \
         Restriction(M, iva_l, dims=dims, dir=dir, dtype=dtype) + \
         Diagonal(weights, dims=dimsd, dir=dir, dtype=dtype) * \
         Restriction(M, iva_r, dims=dims, dir=dir, dtype=dtype)
    return Op, iava


def _sincinterp(M, iava, dims=None, dir=0, dtype='float64'):
    """Sinc interpolation.
    """
    _checkunique(iava)
    # create sinc interpolation matrix
    nreg = M if dims is None else dims[dir]
    ireg = np.arange(nreg)
    sinc = np.tile(iava[:, np.newaxis], (1, nreg)) - \
           np.tile(ireg, (len(iava), 1))
    sinc = np.sinc(sinc)

    # identify additional dimensions and create MatrixMult operator
    otherdims = None
    if dims is not None:
        otherdims = np.array(dims)
        otherdims = np.roll(otherdims, -dir)
        otherdims = otherdims[1:]
        print('dims', dims)
        print('otherdims', otherdims)
    Op = MatrixMult(sinc, dims=otherdims, dtype=dtype)

    # create Transpose operator that brings dir to first dimension
    if dir > 0:
        axes = np.arange(len(dims), dtype=np.int)
        axes = np.roll(axes, -dir)
        dimsd =  list(dims)
        dimsd[dir] = len(iava)
        Top = Transpose(dims, axes=axes, dtype=dtype)
        T1op = Transpose(dimsd, axes=axes, dtype=dtype)
        Op = T1op.H * Op * Top
    return Op


def Interp(M, iava, dims=None, dir=0, kind='linear', dtype='float64'):
    r"""Interpolation operator.

    Apply interpolation along direction ``dir``
    from regularly sampled input vector into fractionary positions ``iava``
    using one of the following algorithms:

    - *Nearest neighbour* interpolation
      is a thin wrapper around :obj:`pylops.Restriction` at ``np.round(iava)``
      locations.

    - *Linear interpolation* extracts values from input vector
      at locations ``np.floor(iava)`` and ``np.floor(iava)+1`` and linearly
      combines them in forward mode, places weighted versions of the
      interpolated values at locations ``np.floor(iava)`` and
      ``np.floor(iava)+1`` in an otherwise zero vector in adjoint mode.

    - *Sinc interpolation* performs sinc interpolation at locations ``iava``.
      Note that this is the most accurate method but it has higher computational
      cost as it involves multiplying the input data by a matrix of size
      :math:`N \times M`.

    .. note:: The vector ``iava`` should contain unique values. If the same
      index is repeated twice an error will be raised. This also applies when
      values beyond the last element of the input array for
      *linear interpolation* as those values are forced to be just before this
      element.

    Parameters
    ----------
    M : :obj:`int`
        Number of samples in model.
    iava : :obj:`list` or :obj:`numpy.ndarray`
         Floating indices of locations of available samples for interpolation.
    dims : :obj:`list`, optional
        Number of samples for each dimension
        (``None`` if only one dimension is available)
    dir : :obj:`int`, optional
        Direction along which restriction is applied.
    kind : :obj:`str`, optional
        Kind of interpolation (``nearest``, ``linear``, and ``sinc`` are
        currently supported)
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Returns
    -------
    op : :obj:`pylops.LinearOperator`
        Linear intepolation operator
    iava : :obj:`list` or :obj:`numpy.ndarray`
        Corrected indices of locations of available samples
        (samples at ``M-1`` or beyond are forced to be at ``M-1-eps``)

    Raises
    ------
    ValueError
        If the vector ``iava`` contains repeated values.
    NotImplementedError
        If ``kind`` is not ``nearest``, ``linear`` or ``sinc``

    See Also
    --------
    pylops.Restriction : Restriction operator

    Notes
    -----
    *Linear interpolation* of a subset of :math:`N` values at locations
    ``iava`` from an input (or model) vector :math:`\mathbf{x}` of size
    :math:`M` can be expressed as:

    .. math::

        y_i = (1-w_i) x_{l^{l}_i} + w_i x_{l^{r}_i}
        \quad \forall i=1,2,...,N

    where :math:`\mathbf{l^l}=[\lfloor l_1 \rfloor, \lfloor l_2 \rfloor,...,
    \lfloor l_N \rfloor]` and :math:`\mathbf{l^r}=[\lfloor l_1 \rfloor +1,
    \lfloor l_2 \rfloor +1,...,
    \lfloor l_N \rfloor +1]` are vectors containing the indeces
    of the original array at which samples are taken, and
    :math:`\mathbf{w}=[l_1 - \lfloor l_1 \rfloor, l_2 - \lfloor l_2 \rfloor,
    ..., l_N - \lfloor l_N \rfloor]` are the linear interpolation weights.
    This operator can be implemented by simply summing two
    :class:`pylops.Restriction` operators which are weighted
    using :class:`pylops.basicoperators.Diagonal` operators.

    *Sinc interpolation* of a subset of :math:`N` values at locations
    ``iava`` from an input (or model) vector :math:`\mathbf{x}` of size
    :math:`M` can be expressed as:

    .. math::

        y_i = \sum_{j=0}^{M} x_j sinc(i-j) \quad \forall i=1,2,...,N

    This operator can be implemented using the :class:`pylops.MatrixMult`
    operator with a matrix containing the values of the sinc function at all
    :math:`i,j` possible combinations.

    """
    if kind == 'nearest':
        interpop, iava = _nearestinterp(M, iava, dims=dims, dir=dir, dtype=dtype)
    elif kind == 'linear':
        interpop, iava = _linearinterp(M, iava, dims=dims, dir=dir, dtype=dtype)
    elif kind == 'sinc':
        interpop = _sincinterp(M, iava, dims=dims, dir=dir, dtype=dtype)
    else:
        raise NotImplementedError('kind is not correct...')
    return LinearOperator(interpop), iava



import numpy as np
from pylops import LinearOperator


class Identity(LinearOperator):
    r"""Identity operator.

    Simply move model to data in forward model and viceversa in adjoint mode if
    :math:`M = N`. If :math:`M > N` removes last :math:`M - N` elements from
    model in forward and pads with :math:`0` in adjoint. If :math:`N > M`
    removes last :math:`N - M` elements from data in adjoint and pads with
    :math:`0` in forward.

    Parameters
    ----------
    N : :obj:`int`
        Number of samples in data (and model, if ``M`` is not provided).
    M : :obj:`int`, optional
        Number of samples in model.
    dtype : :obj:`str`, optional
        Type of elements in input array.
    inplace : :obj:`bool`, optional
        Work inplace (``True``) or make a new copy (``False``). By default,
        data is a reference to the model (in forward) and model is a reference
        to the data (in adjoint).

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    For :math:`M = N`, an *Identity* operator simply moves the model
    :math:`\mathbf{x}` to the data :math:`\mathbf{y}` in forward mode and
    viceversa in adjoint mode:

    .. math::

        y_i = x_i  \quad \forall i=1,2,...,N

    or in matrix form:

    .. math::

        \mathbf{y} = \mathbf{I} \mathbf{x} = \mathbf{x}

    and

    .. math::

        \mathbf{x} = \mathbf{I} \mathbf{y} = \mathbf{y}

    For :math:`M > N`, the *Identity* operator takes the first :math:`M`
    elements of the model :math:`\mathbf{x}` into the data :math:`\mathbf{y}`
    in forward mode

    .. math::

        y_i = x_i  \quad \forall i=1,2,...,N

    and all the elements of the data :math:`\mathbf{y}` into the first
    :math:`M` elements of model in adjoint mode (other elements are ``O``):

    .. math::

        x_i = y_i  \quad \forall i=1,2,...,M

        x_i = 0 \quad \forall i=M+1,...,N

    """
    def __init__(self, N, M=None, dtype='float64', inplace=True):
        M = N if M is None else M
        self.inplace = inplace
        self.shape = (N, M)
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if not self.inplace: x = x.copy()
        if self.shape[0] == self.shape[1]:
            y = x
        elif self.shape[0] < self.shape[1]:
            y = x[:self.shape[0]]
        else:
            y = np.zeros(self.shape[0], dtype=self.dtype)
            y[:self.shape[1]] = x
        return y

    def _rmatvec(self, x):
        if not self.inplace: x = x.copy()
        if self.shape[0] == self.shape[1]:
            y = x
        elif self.shape[0] < self.shape[1]:
            y = np.zeros(self.shape[1], dtype=self.dtype)
            y[:self.shape[0]] = x
        else:
            y = x[:self.shape[1]]
        return y



import numpy as np
import scipy.signal as filt

def _filterdata(d, nt, wav, wcenter):
    r"""Apply filtering to data with wavelet wav
    """
    dwav = filt.lfilter(wav, 1, d, axis=-1)
    dwav = dwav[..., wcenter:]
    d = d[..., :(nt - wcenter)]
    return d, dwav

def makeaxis(par):
    r"""Create axes t, x, and y axes

    Create space and time axes from dictionary containing initial values :math:`(ot, ox, oy)`,
    sampling steps :math:`(dt, dx, dy)` and number of elements :math:`(nt, nx, ny)`
    for each axis

    Parameters
    ----------
    par : :obj:`dict`
        Dictionary containing initial values, sampling steps, and number of elements

    Returns
    -------
    t : :obj:`numpy.ndarray`
        time axis
    t2 : :obj:`numpy.ndarray`
        double time axis (symmetric to zero)
    x : :obj:`numpy.ndarray`
        x axis
    y : :obj:`numpy.ndarray`
        y axis (``None``, if :math:`oy, dy, ny` are not provided)

    Examples
    --------
    >>> par = {'ox':0, 'dx':2, 'nx':60,
    >>>        'oy':0, 'dy':2, 'ny':100,
    >>>        'ot':0, 'dt':4, 'nt':400}
    >>> # Create axis
    >>> t, t2, x, y = makeaxis(par)
    """
    x = par['ox'] + np.arange(par['nx']) * par['dx']
    t = par['ot'] + np.arange(par['nt']) * par['dt']
    t2 = np.arange(-par['nt'] + 1, par['nt']) * par['dt']

    if 'oy' in par.keys():
        y = par['oy'] + np.arange(par['ny']) * par['dy']
    else:
        y = None
    return t, t2, x, y


def linear2d(x, t, v, t0, theta, amp, wav):
    r"""Linear 2D events

    Create 2d linear events given propagation velocity, intercept time, angle,
    and amplitude of each event

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        space axis
    t : :obj:`numpy.ndarray`
        time axis
    v : :obj:`float`
        propagation velocity
    t0 : :obj:`tuple` or :obj:`float`
        intercept time at :math:`x=0` of each linear event
    theta : :obj:`tuple` or :obj:`float`
        angle (in degrees) of each linear event
    amp : :obj:`tuple` or :obj:`float`
        amplitude of each linear event
    wav : :obj:`numpy.ndarray`
        wavelet to be applied to data

    Returns
    -------
    d : :obj:`numpy.ndarray`
        data without wavelet  of size
        :math:`[n_x \times n_t]`
    dwav : :obj:`numpy.ndarray`
        data with wavelet  of size
        :math:`[n_x \times n_t]`

    Notes
    -----
    Each event is created using the following relation:

    .. math::
        t_i(x) = t_{0,i} + p_{x,i} x

    where :math:`p_{x,i}=sin( \theta_i)/v`

    """
    if isinstance(t0, (float, int)):
        t0 = (t0,)
    if isinstance(theta, (float, int)):
        theta = (theta,)
    if isinstance(amp, (float, int)):
        amp = (amp,)

    # identify dimensions
    dt = t[1] - t[0]
    wcenter = int(len(wav)/2)
    nx = np.size(x)
    nt = np.size(t) + wcenter
    nevents = np.size(t0)

    # create events
    d = np.zeros((nx, nt))
    for ievent in range(nevents):
        px = np.sin(np.deg2rad(theta[ievent])) / v
        tevent = t0[ievent] + px * x
        tevent = (tevent - t[0]) / dt
        itevent = tevent.astype(int)
        dtevent = tevent - itevent
        for ix in range(nx):
            if itevent[ix] < nt - 1 and itevent[ix] >= 0:
                d[ix, itevent[ix]] += amp[ievent] * (1 - dtevent[ix])
                d[ix, itevent[ix] + 1] += amp[ievent] * dtevent[ix]

    #filter events with certain wavelet
    d, dwav = _filterdata(d, nt, wav, wcenter)
    return d, dwav


def parabolic2d(x, t, t0, px, pxx, amp, wav):
    r"""Parabolic 2D events

    Create 2d parabolic events given intercept time,
    slowness, curvature, and amplitude of each event

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        space axis
    t : :obj:`numpy.ndarray`
        time axis
    t0 : :obj:`tuple` or :obj:`float`
        intercept time at :math:`x=0` of each parabolic event
    px : :obj:`tuple` or :obj:`float`
        slowness of each parabolic event
    pxx : :obj:`tuple` or :obj:`float`
        curvature of each parabolic event
    amp : :obj:`tuple` or :obj:`float`
        amplitude of each parabolic event
    wav : :obj:`numpy.ndarray`
        wavelet to be applied to data

    Returns
    -------
    d : :obj:`numpy.ndarray`
        data without wavelet of size
        :math:`[n_x \times n_t]`
    dwav : :obj:`numpy.ndarray`
        data with wavelet of size
        :math:`[n_x \times n_t]`

    Notes
    -----
    Each event is created using the following relation:

    .. math::
        t_i(x) = t_{0,i} + p_{x,i} x + p_{xx,i} x^2

    """
    if isinstance(t0, (float, int)):
        t0 = (t0,)
    if isinstance(px, (float, int)):
        px = (px,)
    if isinstance(pxx, (float, int)):
        pxx = (pxx,)
    if isinstance(amp, (float, int)):
        amp = (amp,)

    # identify dimensions
    dt = t[1]-t[0]
    wcenter = int(len(wav)/2)
    nx = np.size(x)
    nt = np.size(t) + wcenter
    nevents = np.size(t0)

    # create events
    d = np.zeros((nx, nt))
    for ievent in range(nevents):
        tevent = t0[ievent] + px[ievent] * x + pxx[ievent] * x ** 2
        tevent = (tevent - t[0]) / dt
        itevent = tevent.astype(int)
        dtevent = tevent - itevent
        for ix in range(nx):
            if itevent[ix] < nt - 1 and itevent[ix] >= 0:
                d[ix, itevent[ix]] += amp[ievent] * (1 - dtevent[ix])
                d[ix, itevent[ix] + 1] += amp[ievent] * dtevent[ix]

    #filter events with certain wavelet
    d, dwav = _filterdata(d, nt, wav, wcenter)
    return d, dwav


def hyperbolic2d(x, t, t0, vrms, amp, wav):
    r"""Hyperbolic 2D events

    Create 2d hyperbolic events given intercept time, root-mean-square
    velocity, and amplitude of each event

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        space axis
    t : :obj:`numpy.ndarray`
        time axis
    t0 : :obj:`tuple` or :obj:`float`
        intercept time at :math:`x=0` of each of hyperbolic event
    vrms : :obj:`tuple` or :obj:`float`
        root-mean-square velocity of each hyperbolic event
    amp : :obj:`tuple` or :obj:`float`
        amplitude of each hyperbolic event
    wav : :obj:`numpy.ndarray`
        wavelet to be applied to data

    Returns
    -------
    d : :obj:`numpy.ndarray`
        data without wavelet of size :math:`[n_x \times n_t]`
    dwav : :obj:`numpy.ndarray`
        data with wavelet of size :math:`[n_x \times n_t]`

    Notes
    -----
    Each event is created using the following relation:

    .. math::
        t_i(x) = \sqrt{t_{0,i}^2 + x^2 / v_{rms,i}^2}

    """
    if isinstance(t0, (float, int)):
        t0 = (t0,)
    if isinstance(vrms, (float, int)):
        vrms = (vrms,)
    if isinstance(amp, (float, int)):
        amp = (amp,)

    # identify dimensions
    dt = t[1]-t[0]
    wcenter = int(len(wav)/2)
    nx = np.size(x)
    nt = np.size(t)+ wcenter
    nevents = np.size(t0)

    #create events
    d = np.zeros((nx, nt))
    for ievent in range(nevents):
        tevent = np.sqrt(t0[ievent] ** 2 + x ** 2 / vrms[ievent] ** 2)
        tevent = (tevent - t[0]) / dt
        itevent = tevent.astype(int)
        dtevent = tevent - itevent
        for ix in range(nx):
            if itevent[ix] < nt - 1 and itevent[ix] >= 0:
                d[ix, itevent[ix]] += amp[ievent] * (1 - dtevent[ix])
                d[ix, itevent[ix] + 1] += amp[ievent] * dtevent[ix]

    #filter events with certain wavelet
    d, dwav = _filterdata(d, nt, wav, wcenter)
    return d, dwav


def linear3d(x, y, t, v, t0, theta, phi, amp, wav):
    r"""Linear 3D events

    Create 3d linear events given propagation velocity, intercept time, angles,
    and amplitude of each event.

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        space axis in x direction
    y : :obj:`numpy.ndarray`
        space axis in y direction
    t : :obj:`numpy.ndarray`
        time axis
    v : :obj:`float`
        propagation velocity
    t0 : :obj:`tuple` or :obj:`float`
        intercept time at :math:`x=0` of each linear event
    theta : :obj:`tuple` or :obj:`float`
        angle in x direction (in degrees) of each linear event
    phi : :obj:`tuple` or :obj:`float`
        angle in y direction (in degrees) of each linear event
    amp : :obj:`tuple` or :obj:`float`
        amplitude of each linear event
    wav : :obj:`numpy.ndarray`
        wavelet to be applied to data

    Returns
    -------
    d : :obj:`numpy.ndarray`
        data without wavelet of size
        :math:`[n_y \times n_x \times n_t]`
    dwav : :obj:`numpy.ndarray`
        data with wavelet of size
        :math:`[n_y \times n_x \times n_t]`

    Notes
    -----
    Each event is created using the following relation:

    .. math::
        t_i(x, y) = t_{0,i} + p_{x,i} x + p_{y,i} y

    where :math:`p_{x,i}=sin( \theta_i)cos( \phi_i)/v`
    and :math:`p_{x,i}=sin( \theta_i)sin( \phi_i)/v`.

    """
    if isinstance(t0, (float, int)):
        t0 = (t0,)
    if isinstance(theta, (float, int)):
        theta = (theta,)
    if isinstance(phi, (float, int)):
        phi = (phi,)
    if isinstance(amp, (float, int)):
        amp = (amp,)

    # identify dimensions
    dt = t[1] - t[0]
    wcenter = int(len(wav)/2)
    nx = np.size(x)
    ny = np.size(y)
    nt = np.size(t) + wcenter
    nevents = np.size(t0)

    #create events
    d = np.zeros((ny, nx, nt))
    for ievent in range(nevents):
        px = np.sin(np.deg2rad(theta[ievent]))*np.cos(np.deg2rad(phi[ievent]))/v
        py = np.sin(np.deg2rad(theta[ievent]))*np.sin(np.deg2rad(phi[ievent]))/v
        for iy in range(ny):
            tevent = t0[ievent] + px * x + py * y[iy]
            tevent = (tevent - t[0]) / dt
            itevent = tevent.astype(int)
            dtevent = tevent - itevent
            for ix in range(nx):
                if itevent[ix] < nt - 1 and itevent[ix] >= 0:
                    d[iy, ix, itevent[ix]] += amp[ievent] * (1 - dtevent[ix])
                    d[iy, ix, itevent[ix] + 1] += amp[ievent] * dtevent[ix]

    #filter events with certain wavelet
    d, dwav = _filterdata(d, nt, wav, wcenter)
    return d, dwav


def hyperbolic3d(x, y, t, t0, vrms_x, vrms_y, amp, wav):
    r"""Hyperbolic 3D events

    Create 3d hyperbolic events given intercept time, root-mean-square
    velocities, and amplitude of each event

    Parameters
    ----------
    x : :obj:`numpy.ndarray`
        space axis in x direction
    y : :obj:`numpy.ndarray`
        space axis in y direction
    t : :obj:`numpy.ndarray`
        time axis
    t0 : :obj:`tuple` or :obj:`float`
        intercept time at :math:`x=0` of each of hyperbolic event
    vrms_x : :obj:`tuple` or :obj:`float`
        root-mean-square velocity in x direction for each hyperbolic event
    vrms_y : :obj:`tuple` or :obj:`float`
        root-mean-square velocity in y direction for each hyperbolic event
    amp : :obj:`tuple` or :obj:`float`
        amplitude of each hyperbolic event
    wav : :obj:`numpy.ndarray`
        wavelet to be applied to data

    Returns
    -------
    d : :obj:`numpy.ndarray`
        data without wavelet of size :math:`[n_y \times n_x \times n_t]`
    dwav : :obj:`numpy.ndarray`
        data with wavelet of size :math:`[n_y \times n_x \times n_t]`

    Notes
    -----
    Each event is created using the following relation:

    .. math::
        t_i(x, y) = \sqrt{t_{0,i}^2 + x^2 / v_{rms_x, i}^2 +
        y^2 / v_{rms_y, i}^2}

    Note that velocities do not have a physical meaning here (compared to the
    corresponding :func:`pylops.utils.seismicevents.hyperbolic2d`), they rather
    simply control the curvature of the hyperboloid along the spatial axes.

    """
    if isinstance(t0, (float, int)):
        t0 = (t0,)
    if isinstance(vrms_x, (float, int)):
        vrms_x = (vrms_x,)
    if isinstance(vrms_y, (float, int)):
        vrms_y = (vrms_y,)
    if isinstance(amp, (float, int)):
        amp = (amp,)

    # identify dimensions
    dt = t[1]-t[0]
    wcenter = int(len(wav)/2)
    nx = np.size(x)
    ny = np.size(y)
    nt = np.size(t) + wcenter
    nevents = np.size(t0)

    #create events
    d = np.zeros((ny, nx, nt))
    for ievent in range(nevents):
        for iy in range(ny):
            tevent = np.sqrt(t0[ievent] ** 2 +
                             x ** 2 / vrms_x[ievent] ** 2 +
                             y[iy] ** 2 / vrms_y[ievent] ** 2)
            tevent = (tevent - t[0]) / dt
            itevent = tevent.astype(int)
            dtevent = tevent - itevent
            for ix in range(nx):
                if itevent[ix] < nt - 1 and itevent[ix] >= 0:
                    d[iy, ix, itevent[ix]] += amp[ievent] * (1 - dtevent[ix])
                    d[iy, ix, itevent[ix] + 1] += amp[ievent] * dtevent[ix]

    #filter events with certain wavelet
    d, dwav = _filterdata(d, nt, wav, wcenter)
    return d, dwav



import logging
import numpy as np

from pylops.basicoperators import Diagonal, BlockDiag, Restriction, \
    HStack
from pylops.utils.tapers import taper3d
from pylops.signalprocessing.Sliding2D import _slidingsteps

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


def Sliding3D(Op, dims, dimsd, nwin, nover, nop,
              tapertype='hanning', design=False):
    """3D Sliding transform operator.

    Apply a transform operator ``Op`` repeatedly to patches of the model
    vector in forward mode and patches of the data vector in adjoint mode.
    More specifically, in forward mode the model vector is divided into patches
    each patch is transformed, and patches are then recombined in a sliding
    window fashion. Both model and data should be 3-dimensional
    arrays in nature as they are internally reshaped and interpreted as
    3-dimensional arrays. Each patch contains in fact a portion of the
    array in the first and second dimensions (and the entire third dimension).

    This operator can be used to perform local, overlapping transforms (e.g.,
    :obj:`pylops.signalprocessing.FFTN`
    or :obj:`pylops.signalprocessing.Radon3D`) of 3-dimensional arrays.

    .. note:: The shape of the model has to be consistent with
       the number of windows for this operator not to return an error. As the
       number of windows depends directly on the choice of ``nwin`` and
       ``nover``, it is recommended to use ``design=True`` if unsure about the
       choice ``dims`` and use the number of windows printed on screen to
       define such input parameter.

    .. warning:: Depending on the choice of `nwin` and `nover` as well as the
       size of the data, sliding windows may not cover the entire first and/or
       second dimensions. The start and end indeces of each window can be
       displayed using ``design=True`` while defining the best sliding window
       approach.

    Parameters
    ----------
    Op : :obj:`pylops.LinearOperator`
        Transform operator
    dims : :obj:`tuple`
        Shape of 3-dimensional model. Note that ``dims[0]`` and ``dims[1]``
        should be multiple of the model sizes of the transform in the
        first and second dimensions
    dimsd : :obj:`tuple`
        Shape of 3-dimensional data
    nwin : :obj:`tuple`
        Number of samples of window
    nover : :obj:`tuple`
        Number of samples of overlapping part of window
    nop : :obj:`tuple`
        Number of samples in axes of transformed domain associated
        to spatial axes in the data
    tapertype : :obj:`str`, optional
        Type of taper (``hanning``, ``cosine``, ``cosinesquare`` or ``None``)
    design : :obj:`bool`, optional
        Print number sliding window (``True``) or not (``False``)

    Returns
    -------
    Sop : :obj:`pylops.LinearOperator`
        Sliding operator

    Raises
    ------
    ValueError
        Identified number of windows is not consistent with provided model
        shape (``dims``).

    """
    # model windows
    mwin0_ins, mwin0_ends = _slidingsteps(dims[0],
                                          Op.shape[1]//(nop[1]*dims[2]), 0)
    mwin1_ins, mwin1_ends = _slidingsteps(dims[1],
                                          Op.shape[1]//(nop[0]*dims[2]), 0)

    # data windows
    dwin0_ins, dwin0_ends = _slidingsteps(dimsd[0], nwin[0], nover[0])
    dwin1_ins, dwin1_ends = _slidingsteps(dimsd[1], nwin[1], nover[1])
    nwins0 = len(dwin0_ins)
    nwins1 = len(dwin1_ins)
    nwins = nwins0*nwins1

    # create tapers
    if tapertype is not None:
        tap = taper3d(dimsd[2], nwin, nover, tapertype=tapertype)

    # check that identified number of windows agrees with mode size
    if design:
        logging.warning('(%d,%d) windows required...', nwins0, nwins1)
        logging.warning('model wins - start0:%s, end0:%s, start1:%s, end1:%s',
                        str(mwin0_ins), str(mwin0_ends),
                        str(mwin1_ins), str(mwin1_ends))
        logging.warning('data wins - start0:%s, end0:%s, start1:%s, end1:%s',
                        str(dwin0_ins), str(dwin0_ends),
                        str(dwin1_ins), str(dwin1_ends))

    if nwins*Op.shape[1]//dims[2] != dims[0]*dims[1]:
        raise ValueError('Model shape (dims=%s) is not consistent with chosen '
                         'number of windows. Choose dims[0]=%d and '
                         'dims[1]=%d for the operator to work with '
                         'estimated number of windows, or create '
                         'the operator with design=True to find out the'
                         'optimal number of windows for the current '
                         'model size...'
                         % (str(dims), nwins0*Op.shape[1]//(nop[1]*dims[2]),
                            nwins1 * Op.shape[1]//(nop[0]*dims[2])))
    # transform to apply
    if tapertype is None:
        OOp = BlockDiag([Op for _ in range(nwins)])
    else:
        OOp = BlockDiag([Diagonal(tap.flatten()) * Op
                         for _ in range(nwins)])

    hstack = HStack([Restriction(dimsd[1] * dimsd[2] * nwin[0],
                                 range(win_in, win_end),
                                 dims=(nwin[0], dimsd[1], dimsd[2]),
                                 dir=1).H
                     for win_in, win_end in zip(dwin1_ins,
                                                dwin1_ends)])

    combining1 = BlockDiag([hstack]*nwins0)
    combining0 = HStack([Restriction(np.prod(dimsd),
                                     range(win_in, win_end),
                                     dims=dimsd, dir=0).H
                         for win_in, win_end in zip(dwin0_ins, dwin0_ends)])
    Sop = combining0 * combining1 * OOp
    return Sop



import numpy as np
from pylops import LinearOperator


class Diagonal(LinearOperator):
    r"""Diagonal operator.

    Applies element-wise multiplication of the input vector with the vector
    ``diag`` in forward and with its complex conjugate in adjoint mode.

    This operator can also broadcast; in this case the input vector is
    reshaped into its dimensions ``dims`` and the element-wise multiplication
    with ``diag`` is perfomed on the direction ``dir``. Note that the
    vector ``diag`` will need to have size equal to ``dims[dir]``.

    Parameters
    ----------
    diag : :obj:`numpy.ndarray`
        Vector to be used for element-wise multiplication.
    dims : :obj:`list`, optional
        Number of samples for each dimension
        (``None`` if only one dimension is available)
    dir : :obj:`int`, optional
        Direction along which multiplication is applied.
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    Element-wise multiplication between the model :math:`\mathbf{x}` and/or
    data :math:`\mathbf{y}` vectors and the array :math:`\mathbf{d}`
    can be expressed as

    .. math::

        y_i = d_i x_i  \quad \forall i=1,2,...,N

    This is equivalent to a matrix-vector multiplication with a matrix
    containing the vector :math:`\mathbf{d}` along its main diagonal.

    For real-valued ``diag``, the Diagonal operator is self-adjoint as the
    adjoint of a diagonal matrix is the diagonal matrix itself. For
    complex-valued ``diag``, the adjoint is equivalent to the element-wise
    multiplication with the complex conjugate elements of ``diag``.

    """
    def __init__(self, diag, dims=None, dir=0, dtype='float64'):
        self.diag = diag.flatten()
        self.complex = True if np.iscomplexobj(self.diag) else False
        if dims is None:
            self.shape = (len(self.diag), len(self.diag))
            self.dims = None
            self.reshape = False
        else:
            diagdims = [1] * len(dims)
            diagdims[dir] = dims[dir]
            self.diag = self.diag.reshape(diagdims)
            self.shape = (np.prod(dims), np.prod(dims))
            self.dims = dims
            self.reshape = True
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        if not self.reshape:
            y = self.diag * x.ravel()
        else:
            x = x.reshape(self.dims)
            y = self.diag * x
        return y.ravel()

    def _rmatvec(self, x):
        if self.complex:
            diagadj = self.diag.conj()
        else:
            diagadj = self.diag
        if not self.reshape:
            y = diagadj * x.ravel()
        else:
            x = x.reshape(self.dims)
            y = diagadj * x
        return y.ravel()

    def matrix(self):
        """Return diagonal matrix as dense :obj:`numpy.ndarray`

        Returns
        ----------
        densemat : :obj:`numpy.ndarray`
            Dense matrix.

        """
        densemat = np.diag(self.diag.squeeze())
        return densemat



import numpy as np
from pylops.basicoperators import FirstDerivative, VStack


def Gradient(dims, sampling=1, edge=False, dtype='float64', kind='centered'):
    r"""Gradient.

    Apply gradient operator to a multi-dimensional
    array (at least 2 dimensions are required).

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension.
    sampling : :obj:`tuple`, optional
        Sampling steps for each direction.
    edge : :obj:`bool`, optional
        Use reduced order derivative at edges (``True``) or
        ignore them (``False``).
    dtype : :obj:`str`, optional
        Type of elements in input array.
    kind : :obj:`str`, optional
        Derivative kind (``forward``, ``centered``, or ``backward``).

    Returns
    -------
    l2op : :obj:`pylops.LinearOperator`
        Gradient linear operator

    Notes
    -----
    The Gradient operator applies a first-order derivative to each dimension of
    a multi-dimensional array in forward mode.

    For simplicity, given a three dimensional array, the Gradient in forward
    mode using a centered stencil can be expressed as:

    .. math::
        \mathbf{g}_{i, j, k} =
            (f_{i+1, j, k} - f_{i-1, j, k}) / d_1 \mathbf{i_1} +
            (f_{i, j+1, k} - f_{i, j-1, k}) / d_2 \mathbf{i_2} +
            (f_{i, j, k+1} - f_{i, j, k-1}) / d_3 \mathbf{i_3}

    which is discretized as follows:

    .. math::
        \mathbf{g}  =
        \begin{bmatrix}
           \mathbf{df_1} \\
           \mathbf{df_2} \\
           \mathbf{df_3}
        \end{bmatrix}

    In adjoint mode, the adjoints of the first derivatives along different
    axes are instead summed together.

    """
    ndims = len(dims)
    if isinstance(sampling, (int, float)):
        sampling = [sampling] * ndims

    gop = VStack([FirstDerivative(np.prod(dims), dims=dims, dir=idir,
                                  sampling=sampling[idir],
                                  edge=edge, kind=kind, dtype=dtype)
                  for idir in range(ndims)])
    return gop



import numpy as np
from pylops import LinearOperator


class FFT2D(LinearOperator):
    r"""Two dimensional Fast-Fourier Transform.

    Apply two dimensional Fast-Fourier Transform (FFT) to any pair of axes of a
    multi-dimensional array depending on the choice of ``dirs``.
    Note that the FFT2D operator is a simple overload to the numpy
    :py:func:`numpy.fft.fft2` in forward mode and to the numpy
    :py:func:`numpy.fft.ifft2` in adjoint mode, however scaling is taken
    into account differently to guarantee that the operator is passing the
    dot-test.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension
    dirs : :obj:`tuple`, optional
        Pair of directions along which FFT2D is applied
    nffts : :obj:`tuple`, optional
        Number of samples in Fourier Transform for each direction (same as
        input if ``nffts=(None, None)``)
    sampling : :obj:`tuple`, optional
        Sampling steps ``dy`` and ``dx``
    dtype : :obj:`str`, optional
        Type of elements in input array. Note that the `dtype` of the operator
        is the corresponding complex type even when a real type is provided.
        Nevertheless, the provided dtype will be enforced on the vector
        returned by the `rmatvec` method.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (True) or not (False)

    Raises
    ------
    ValueError
        If ``dims`` has less than two elements, and if ``dirs``, ``nffts``,
        or ``sampling`` has more or less than two elements.

    Notes
    -----
    The FFT2D operator applies the two-dimensional forward Fourier transform
    to a signal :math:`d(y,x)` in forward mode:

    .. math::
        D(k_y, k_x) = \mathscr{F} (d) = \int \int d(y,x) e^{-j2\pi k_yy}
        e^{-j2\pi k_xx} dy dx

    Similarly, the  two-dimensional inverse Fourier transform is applied to
    the Fourier spectrum :math:`D(k_y, k_x)` in adjoint mode:

    .. math::
        d(y,x) = \mathscr{F}^{-1} (D) = \int \int D(k_y, k_x) e^{j2\pi k_yy}
        e^{j2\pi k_xx} dk_y  dk_x

    Both operators are effectively discretized and solved by a fast iterative
    algorithm known as Fast Fourier Transform.

    """
    def __init__(self, dims, dirs=(0, 1), nffts=(None, None),
                 sampling=(1., 1.), dtype='complex128'):
        # checks
        if len(dims) < 2:
            raise ValueError('provide at least two dimensions')
        if len(dirs) != 2:
            raise ValueError('provide at two directions along which fft is applied')
        if len(nffts) != 2:
            raise ValueError('provide at two nfft dimensions')
        if len(sampling) != 2:
            raise ValueError('provide two sampling steps')

        self.dirs = dirs
        self.nffts = np.array([int(nffts[0]) if nffts[0] is not None
                               else dims[self.dirs[0]],
                               int(nffts[1]) if nffts[1] is not None
                               else dims[self.dirs[1]]])
        self.f1 = np.fft.fftfreq(self.nffts[0], d=sampling[0])
        self.f2 = np.fft.fftfreq(self.nffts[1], d=sampling[1])

        self.dims = np.array(dims)
        self.dims_fft = self.dims.copy()
        self.dims_fft[self.dirs[0]] = self.nffts[0]
        self.dims_fft[self.dirs[1]] = self.nffts[1]

        self.shape = (int(np.prod(self.dims_fft)), int(np.prod(self.dims)))
        self.rdtype = np.dtype(dtype)
        self.cdtype = (np.ones(1, dtype=self.rdtype) +
                       1j * np.ones(1, dtype=self.rdtype)).dtype
        self.dtype = self.cdtype
        self.explicit = False

    def _matvec(self, x):
        x = np.reshape(x, self.dims)
        y = np.sqrt(1./np.prod(self.nffts)) * np.fft.fft2(x, s=self.nffts,
                                                          axes=(self.dirs[0],
                                                                self.dirs[1]))
        y = y.flatten()
        y = y.astype(self.cdtype)
        return y

    def _rmatvec(self, x):
        x = np.reshape(x, self.dims_fft)
        y = np.sqrt(np.prod(self.nffts)) * np.fft.ifft2(x, s=self.nffts,
                                                        axes=(self.dirs[0],
                                                              self.dirs[1]))
        y = np.take(y, range(self.dims[self.dirs[0]]), axis=self.dirs[0])
        y = np.take(y, range(self.dims[self.dirs[1]]), axis=self.dirs[1])
        y = y.flatten()
        y = y.astype(self.rdtype)
        return y
    


import numpy as np


def _chirp_radon_2d(data, dt, dx, pmax, mode='f'):
    r"""2D Chirp Radon transform

    Applies 2D Radon transform using Fast Fourier Transform and Chirp
    functions. (mode='f': forward, 'a': adjoint, and  'i': inverse). See
    Chirp2DRadon operator docstring for more details.

    Parameters
    ----------
    data : :obj:`np.ndarray`
        2D input data of size :math:`[n_x \times n_t]`
    dt : :obj:`float`
        Time sampling :math:`dt`
    dx : :obj:`float`
        Spatial sampling in :math:`x` direction :math:`dx`
    pmax : :obj:`np.ndarray`
        Maximum slope defined as :math:`\tan` of maximum stacking angle
        :math:`x` direction :math:`p_{max} = \tan(\alpha_x_{max})`.
        If one operates in terms of minimum velocity :math:`c_0`, then
        :math:`p_x_{max}=c_0 dy/dt`.
    mode : :obj:`str`, optional
        Mode of operation, 'f': forward, 'a': adjoint, and  'i': inverse

    Returns
    -------
    g : :obj:`np.ndarray`
        2D output of size :math:`[\times n_{x} \times n_t]`

    """
    # define sign for mode
    sign = -1. if mode == 'f' else 1.

    # data size
    (nx, nt) = data.shape

    # find dtype of input
    dtype = np.real(data).dtype
    cdtype = (np.ones(1, dtype=dtype) + 1j * np.ones(1, dtype=dtype)).dtype

    # frequency axis
    omega = (np.fft.fftfreq(nt, 1 / nt) / (nt*dt)).reshape((1, nt)).astype(dtype)

    # slowness sampling
    dp = 2 * dt * pmax / dx / nx

    # spatial axis
    x = (np.fft.fftfreq(2*nx, 1 / (2*nx)) ** 2).reshape((2*nx, 1)).astype(dtype)

    # K coefficients
    K0 = np.exp(sign * np.pi * 1j * dp * dx * omega * x).reshape((2*nx, nt))

    # K conj coefficients
    K = np.conj(np.fft.fftshift(K0, axes=(0,)))[nx//2:3*nx//2, :]

    # perform transform
    h = np.zeros((2 * nx, nt)).astype(cdtype)
    h[0:nx, :] = np.fft.fftn(data, axes=(1,)) * K
    g = np.fft.ifftn(np.fft.fftn(h, axes=(0,)) *
                     np.fft.fftn(K0, axes=(0,)), axes=(0,))
    if mode == 'i':
        g = np.fft.ifftn(g[0:nx, :] * K * abs(omega), axes=(1,)).real * dp * dx
    else:
        g = np.fft.ifftn(g[0:nx, :] * K, axes=(1,)).real
    return g



import numpy as np
from scipy.signal import convolve, correlate
from pylops import LinearOperator


class ConvolveND(LinearOperator):
    r"""ND convolution operator.

    Apply n-dimensional convolution with a compact filter to model
    (and data) along a set of directions ``dirs`` of a n-dimensional
    array.

    Parameters
    ----------
    N : :obj:`int`
        Number of samples in model
    h : :obj:`numpy.ndarray`
        nd compact filter to be convolved to input signal
    dims : :obj:`list`
        Number of samples for each dimension
    offset : :obj:`tuple`, optional
        Indices of the center of the compact filter
    dirs : :obj:`tuple`, optional
        Directions along which convolution is applied
        (set to ``None`` for filter of same dimension as input vector)
    method : :obj:`str`, optional
        Method used to calculate the convolution (``direct`` or ``fft``).
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved
        explicitly (``True``) or not (``False``)

    Notes
    -----
    The ConvolveND operator applies n-dimensional convolution
    between the input signal :math:`d(x_1, x_2, ..., x_N)` and a compact
    filter kernel :math:`h(x_1, x_2, ..., x_N)` in forward model. This is a
    straighforward extension to multiple dimensions of
    :obj:`pylops.signalprocessing.Convolve2D` operator.

    """
    def __init__(self, N, h, dims, offset=(0, 0, 0), dirs=None,
                 method='fft', dtype='float64'):
        self.h = np.array(h)
        self.nh = np.array(self.h.shape)
        self.dirs = np.arange(len(dims)) if dirs is None else np.array(dirs)

        # find out which directions are used for convolution and define offsets
        if len(dims) != len(self.nh):
            self.offset = self.nh // 2
            dimsh = np.ones(len(dims), dtype=np.int)
            for dir in self.dirs:
                dimsh[dir] = self.nh[dir]
                self.offset[dir] = int(offset[dir])
            self.h = self.h.reshape(dimsh)
        else:
            self.offset = np.array(offset).astype(np.int)
        for dir in self.dirs:
            self.offset[dir] = int(offset[dir])

        # padding
        self.offset = 2 * (self.nh // 2 - self.offset)
        pad = [(0, 0) for _ in range(len(dims))]
        dopad = False
        for inh, nh in enumerate(self.nh):
            if nh % 2 == 0:
                self.offset[inh] -= 1
            if self.offset[inh] != 0:
                pad[inh] = [self.offset[inh] if self.offset[inh] > 0 else 0,
                            -self.offset[inh] if self.offset[inh] < 0 else 0]
                dopad = True
        if dopad:
            self.h = np.pad(self.h, pad, mode='constant')

        if np.prod(dims) != N:
            raise ValueError('product of dims must equal N!')
        else:
            self.dims = np.array(dims)
            self.reshape = True
        self.shape = (np.prod(self.dims), np.prod(self.dims))
        self.method = method
        self.dtype = np.dtype(dtype)
        self.explicit = False

    def _matvec(self, x):
        x = np.reshape(x, self.dims)
        y = convolve(x, self.h, mode='same', method=self.method)
        y = y.ravel()
        return y

    def _rmatvec(self, x):
        x = np.reshape(x, self.dims)
        y = correlate(x, self.h, mode='same', method=self.method)
        y = y.ravel()
        return y



import logging
import numpy as np
from pylops import LinearOperator

try:
    import pyfftw
except ModuleNotFoundError:
    pyfftw = None
    pyfftw_message = 'Pyfftw not installed, use numpy or run ' \
                     '"pip install pyFFTW" or ' \
                     '"conda install -c conda-forge pyfftw".'
except Exception as e:
    pyfftw = None
    pyfftw_message = 'Failed to import pyfftw (error:%s), use numpy.' % e

logging.basicConfig(format='%(levelname)s: %(message)s', level=logging.WARNING)


class _FFT_numpy(LinearOperator):
    """One dimensional Fast-Fourier Transform using numpy
    """
    def __init__(self, dims, dir=0, nfft=None, sampling=1.,
                 real=False, fftshift=False, dtype='complex128'):
        if isinstance(dims, int):
            dims = (dims,)
        if dir > len(dims)-1:
            raise ValueError('dir=%d must be smaller than '
                             'number of dims=%d...' % (dir, len(dims)))
        self.dir = dir
        self.nfft = nfft if nfft is not None else dims[self.dir]
        self.real = real
        self.fftshift = fftshift
        self.f = np.fft.rfftfreq(self.nfft, d=sampling) if real \
                 else np.fft.fftfreq(self.nfft, d=sampling)
        if len(dims) == 1:
            self.dims = np.array([dims[0], 1])
            self.dims_fft = self.dims.copy()
            self.dims_fft[self.dir] = self.nfft // 2 + 1 if \
                self.real else self.nfft
            self.reshape = False
        else:
            self.dims = np.array(dims)
            self.dims_fft = self.dims.copy()
            self.dims_fft[self.dir] = self.nfft // 2 + 1 if \
                self.real else self.nfft
            self.reshape = True
        self.shape = (int(np.prod(self.dims_fft)), int(np.prod(self.dims)))
        # Find types to enforce to forward and adjoint outputs. This is
        # required as np.fft.fft always returns complex128 even if input is
        # float32 or less. Moreover, when choosing real=True, the type of the
        # adjoint output is forced to be real even if the provided dtype
        # is complex.
        self.rdtype = np.real(np.ones(1, dtype)).dtype if real else np.dtype(dtype)
        self.cdtype = (np.ones(1, dtype=self.rdtype) +
                       1j * np.ones(1, dtype=self.rdtype)).dtype
        self.dtype = self.cdtype
        self.explicit = False

    def _matvec(self, x):
        if not self.reshape:
            x = x.ravel()
            if self.fftshift:
                x = np.fft.ifftshift(x)
            if self.real:
                y = np.sqrt(1./self.nfft)*np.fft.rfft(np.real(x),
                                                      n=self.nfft, axis=-1)
            else:
                y = np.sqrt(1./self.nfft)*np.fft.fft(x, n=self.nfft, axis=-1)
        else:
            x = np.reshape(x, self.dims)
            if self.fftshift:
                x = np.fft.ifftshift(x, axes=self.dir)
            if self.real:
                y = np.sqrt(1. / self.nfft) * np.fft.rfft(np.real(x),
                                                          n=self.nfft,
                                                          axis=self.dir)
            else:
                y = np.sqrt(1. / self.nfft) * np.fft.fft(x, n=self.nfft,
                                                         axis=self.dir)
            y = y.flatten()
        y = y.astype(self.cdtype)
        return y

    def _rmatvec(self, x):
        if not self.reshape:
            x = x.ravel()
            if self.real:
                y = np.sqrt(self.nfft)*np.fft.irfft(x, n=self.nfft, axis=-1)
                y = np.real(y)
            else:
                y = np.sqrt(self.nfft)*np.fft.ifft(x, n=self.nfft, axis=-1)
            if self.nfft != self.dims[self.dir]:
                y = y[:self.dims[self.dir]]
            if self.fftshift:
                y = np.fft.fftshift(y)
        else:
            x = np.reshape(x, self.dims_fft)
            if self.real:
                y = np.sqrt(self.nfft) * np.fft.irfft(x, n=self.nfft,
                                                      axis=self.dir)
                y = np.real(y)
            else:
                y = np.sqrt(self.nfft) * np.fft.ifft(x, n=self.nfft,
                                                     axis=self.dir)
            if self.nfft != self.dims[self.dir]:
                y = np.take(y, np.arange(0, self.dims[self.dir]),
                            axis=self.dir)
            if self.fftshift:
                y = np.fft.fftshift(y, axes=self.dir)
            y = y.flatten()
        y = y.astype(self.rdtype)
        return y


class _FFT_fftw(LinearOperator):
    """One dimensional Fast-Fourier Transform using pyffw
    """
    def __init__(self, dims, dir=0, nfft=None, sampling=1.,
                 real=False, fftshift=False, dtype='complex128',
                 **kwargs_fftw):
        if isinstance(dims, int):
            dims = (dims,)
        if dir > len(dims)-1:
            raise ValueError('dir=%d must be smaller than '
                             'number of dims=%d...' % (dir, len(dims)))
        self.dir = dir
        if nfft is None:
            nfft = dims[self.dir]
        elif nfft < dims[self.dir]:
            logging.warning('nfft should be bigger or equal then '
                            ' dims[self.dir] for engine=fftw, set to '
                            'dims[self.dir]')
            nfft = dims[self.dir]
        self.nfft = nfft

        self.real = real
        self.fftshift = fftshift
        self.f = np.fft.rfftfreq(self.nfft, d=sampling) if real \
                 else np.fft.fftfreq(self.nfft, d=sampling)
        if len(dims) == 1:
            self.dims = np.array([dims[0], ])
            self.dims_t = self.dims.copy()
            self.dims_t[self.dir] = self.nfft
            self.dims_fft = self.dims.copy()
            self.dims_fft[self.dir] = self.nfft // 2 + 1 if \
                self.real else self.nfft
            self.reshape = False
        else:
            self.dims = np.array(dims)
            self.dims_t = self.dims.copy()
            self.dims_t[self.dir] = self.nfft
            self.dims_fft = self.dims.copy()
            self.dims_fft[self.dir] = self.nfft//2 + 1 if \
                self.real else self.nfft
            self.reshape = True
        self.shape = (int(np.prod(self.dims_fft)), int(np.prod(self.dims)))
        self.rdtype = np.real(np.ones(1, dtype)).dtype if real else np.dtype(dtype)
        self.cdtype = (np.ones(1, dtype=self.rdtype) +
                       1j * np.ones(1, dtype=self.rdtype)).dtype
        self.dtype = self.cdtype
        self.explicit = False

        # define padding(fftw requires the user to provide padded input signal)
        self.pad = [[0, 0] for _ in range(len(self.dims))]
        if self.real:
            if self.nfft % 2:
                self.pad[self.dir][1] = 2*(self.dims_fft[self.dir]-1) + 1 - \
                                        self.dims[self.dir]
            else:
                self.pad[self.dir][1] = 2*(self.dims_fft[self.dir]-1) - \
                                        self.dims[self.dir]
        else:
            self.pad[self.dir][1] = self.dims_fft[self.dir] - \
                                      self.dims[self.dir]
        self.dopad = True if np.sum(np.array(self.pad)) > 0 else False

        # create empty arrays and plans for fft/ifft
        self.x = pyfftw.empty_aligned(self.dims_t,
                                      dtype=self.rdtype if real else self.cdtype)
        self.y = pyfftw.empty_aligned(self.dims_fft,
                                      dtype=self.cdtype)
        self.fftplan = pyfftw.FFTW(self.x, self.y, axes=(self.dir,),
                                   direction='FFTW_FORWARD', **kwargs_fftw)
        self.ifftplan = pyfftw.FFTW(self.y, self.x, axes=(self.dir,),
                                    direction='FFTW_BACKWARD', **kwargs_fftw)

    def _matvec(self, x):
        if self.real:
            x = np.real(x)
        if not self.reshape:
            x = x.ravel()
            if self.fftshift:
                x = np.fft.ifftshift(x)
            if self.dopad:
                x = np.pad(x, self.pad, 'constant', constant_values=0)

            y = np.sqrt(1./self.nfft)*self.fftplan(x)
        else:
            x = np.reshape(x, self.dims)
            if self.fftshift:
                x = np.fft.ifftshift(x, axes=self.dir)
            if self.dopad:
                x = np.pad(x, self.pad, 'constant', constant_values=0)
            y = np.sqrt(1. / self.nfft) * self.fftplan(x)
            y = np.ndarray.flatten(y)
        return y

    def _rmatvec(self, x):
        if not self.reshape:
            x = x.ravel()
            y = np.sqrt(self.nfft) * self.ifftplan(x)
            if self.nfft != self.dims[self.dir]:
                y = y[:self.dims[self.dir]]
            if self.fftshift:
                y = np.fft.fftshift(y)
        else:
            x = np.reshape(x, self.dims_fft)
            y = np.sqrt(self.nfft) * self.ifftplan(x)
            if self.nfft != self.dims[self.dir]:
                y = np.take(y, np.arange(0, self.dims[self.dir]),
                            axis=self.dir)
            if self.fftshift:
                y = np.fft.fftshift(y, axes=self.dir)
            y = np.ndarray.flatten(y)
        if self.real:
            y = np.real(y)
        return y


def FFT(dims, dir=0, nfft=None, sampling=1., real=False,
        fftshift=False, engine='numpy', dtype='complex128', **kwargs_fftw):
    r"""One dimensional Fast-Fourier Transform.

    Apply Fast-Fourier Transform (FFT) along a specific direction ``dir`` of a
    multi-dimensional array of size ``dim``.

    Note that the FFT operator is an overload to either the numpy
    :py:func:`numpy.fft.fft` (or :py:func:`numpy.fft.rfft` for real models) in
    forward mode and to the numpy :py:func:`numpy.fft.ifft` (or
    :py:func:`numpy.fft.irfft` for real models) in adjoint mode, or to the
    :py:class:`pyfftw.FFTW` class.

    In both cases, scaling is properly taken into account to guarantee
    that the operator is passing the dot-test.

    .. note:: For a real valued input signal, it is possible to store the
      values of the Fourier transform at positive frequencies only as values
      at negative frequencies are simply their complex conjugates.
      However as the operation of removing the negative part of the frequency
      axis in forward mode and adding the complex conjugates in adjoint mode
      is nonlinear, the Linear Operator FTT with ``real=True`` is not expected
      to pass the dot-test. It is thus *only* advised to use this flag when a
      forward and adjoint FFT is used in the same chained operator
      (e.g., ``FFT.H*Op*FFT``) such as in
      :py:func:`pylops.waveeqprocessing.mdd.MDC`.

    Parameters
    ----------
    dims : :obj:`tuple`
        Number of samples for each dimension
    dir : :obj:`int`, optional
        Direction along which FFT is applied.
    nfft : :obj:`int`, optional
        Number of samples in Fourier Transform (same as input if ``nfft=None``)
    sampling : :obj:`float`, optional
        Sampling step ``dt``.
    real : :obj:`bool`, optional
        Model to which fft is applied has real numbers (``True``) or not
        (``False``). Used to enforce that the output of adjoint of a real
        model is real.
    fftshift : :obj:`bool`, optional
        Apply fftshift/ifftshift (``True``) or not (``False``)
    engine : :obj:`str`, optional
        Engine used for fft computation (``numpy`` or ``fftw``)
    dtype : :obj:`str`, optional
        Type of elements in input array. Note that the `dtype` of the operator
        is the corresponding complex type even when a real type is provided.
        Nevertheless, the provided dtype will be enforced on the vector
        returned by the `rmatvec` method.
    **kwargs_fftw
            Arbitrary keyword arguments
            for :py:class:`pyfftw.FTTW`

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly
        (True) or not (False)

    Raises
    ------
    ValueError
        If ``dims`` is not provided and if ``dir`` is bigger than ``len(dims)``
    NotImplementedError
        If ``engine`` is neither ``numpy`` nor ``numba``

    Notes
    -----
    The FFT operator applies the forward Fourier transform to a signal
    :math:`d(t)` in forward mode:

    .. math::
        D(f) = \mathscr{F} (d) = \int d(t) e^{-j2\pi ft} dt

    Similarly, the inverse Fourier transform is applied to the Fourier spectrum
    :math:`D(f)` in adjoint mode:

    .. math::
        d(t) = \mathscr{F}^{-1} (D) = \int D(f) e^{j2\pi ft} df

    Both operators are effectively discretized and solved by a fast iterative
    algorithm known as Fast Fourier Transform.

    Note that the FFT operator is a special operator in that the adjoint is
    also the inverse of the forward mode. Moreover, in case of real signal
    in time domain, the Fourier transform in Hermitian.

    """
    if engine == 'fftw' and pyfftw is not None:
        f = _FFT_fftw(dims, dir=dir, nfft=nfft, sampling=sampling,
                      real=real, fftshift=fftshift, dtype=dtype, **kwargs_fftw)
    elif engine == 'numpy' or (engine == 'fftw' and pyfftw is None):
        if engine == 'fftw' and pyfftw is None:
            logging.warning(pyfftw_message)
        f = _FFT_numpy(dims, dir=dir, nfft=nfft, sampling=sampling,
                       real=real, fftshift=fftshift, dtype=dtype)
    else:
        raise NotImplementedError('engine must be numpy or fftw')
    return f



import numpy as np
from pylops.signalprocessing import Convolve1D


def Smoothing1D(nsmooth, dims, dir=0, dtype='float64'):
    r"""1D Smoothing.

    Apply smoothing to model (and data) along a specific direction of a
    multi-dimensional array depending on the choice of ``dir``.

    Parameters
    ----------
    nsmooth : :obj:`int`
        Lenght of smoothing operator (must be odd)
    dims : :obj:`tuple` or :obj:`int`
        Number of samples for each dimension
    dir : :obj:`int`, optional
        Direction along which smoothing is applied
    dtype : :obj:`str`, optional
        Type of elements in input array.

    Attributes
    ----------
    shape : :obj:`tuple`
        Operator shape
    explicit : :obj:`bool`
        Operator contains a matrix that can be solved explicitly (``True``) or
        not (``False``)

    Notes
    -----
    The Smoothing1D operator is a special type of convolutional operator that
    convolves the input model (or data) with a constant filter of size
    :math:`n_{smooth}`:

    .. math::
        \mathbf{f} = [ 1/n_{smooth}, 1/n_{smooth}, ..., 1/n_{smooth} ]

    When applied to the first direction:

    .. math::
        y[i,j,k] = 1/n_{smooth} \sum_{l=-(n_{smooth}-1)/2}^{(n_{smooth}-1)/2}
        x[l,j,k]

    Similarly when applied to the second direction:

    .. math::
        y[i,j,k] = 1/n_{smooth} \sum_{l=-(n_{smooth}-1)/2}^{(n_{smooth}-1)/2}
        x[i,l,k]

    and the third direction:

    .. math::
        y[i,j,k] = 1/n_{smooth} \sum_{l=-(n_{smooth}-1)/2}^{(n_{smooth}-1)/2}
        x[i,j,l]

    Note that since the filter is symmetrical, the *Smoothing1D* operator is
    self-adjoint.

    """
    if isinstance(dims, int):
        dims = (dims,)
    if nsmooth % 2 == 0:
        nsmooth += 1

    return Convolve1D(np.prod(np.array(dims)), dims=dims, dir=dir,
                      h=np.ones(nsmooth)/float(nsmooth), offset=(nsmooth-1)/2,
                      dtype=dtype)
